from typing import List, Union, Optional, Generator, Set, Type

from merakicommons.cache import lazy_property, lazy
from merakicommons.container import searchable, SearchableList

from .. import configuration
from ..data import Region, Platform, Tier, Division, Queue, Role
from .common import CoreData, CoreDataList, CassiopeiaObject, CassiopeiaGhost, CassiopeiaLazyList, provide_default_region, ghost_load_on
from ..dto.league import LeagueDto, LeagueEntryDto, LeagueEntriesDto, LeagueSummonerEntriesDto, MiniSeriesDto, GrandmasterLeagueListDto, ChallengerLeagueListDto, MasterLeagueListDto
from .summoner import Summoner


##############
# Data Types #
##############


class MiniSeriesData(CoreData):
    _dto_type = MiniSeriesDto
    _renamed = {}


class LeagueEntryData(CoreData):
    """Contains the data for one entry (summoner) in a League."""
    _dto_type = LeagueEntryDto
    _renamed = {"miniSeries": "promos", "playerOrTeamId": "summonerId", "playerOrTeamName": "summonerName", "leagueName": "name", "queueType": "queue", "rank": "division"}

    def __call__(self, **kwargs):
        if "miniSeries" in kwargs:
            self.promos = MiniSeriesData(**kwargs.pop("miniSeries"))
        if "summonerId" in kwargs:
            self.summonerId = kwargs.pop("summonerId")
        super().__call__(**kwargs)
        return self


class LeagueData(CoreData):
    """Contains the data for one League which has many entries."""
    _dto_type = LeagueDto
    _renamed = {"leagueId": "id"}

    def __call__(self, **kwargs):
        if "entries" in kwargs:
            self.entries = [LeagueEntryData(**entry) for entry in kwargs.pop("entries")]
        super().__call__(**kwargs)
        return self


class LeagueSummonerEntriesData(CoreDataList):
    """League entries for a single summoner."""
    _dto_type = LeagueSummonerEntriesDto
    _renamed = {"summonerId": "summoner_id"}


class LeagueEntriesData(CoreDataList):
    """For paginated League entries."""
    _dto_type = LeagueEntriesDto
    _renamed = {}


class ChallengerLeagueListData(CoreData):
    _dto_type = ChallengerLeagueListDto
    _renamed = {"leagueId": "id"}

    def __call__(self, **kwargs):
        if "entries" in kwargs:
            self.entries = [LeagueEntryData(**entry) for entry in kwargs.pop("entries")]
        super().__call__(**kwargs)
        return self

class GrandmasterLeagueListData(CoreData):
    _dto_type = GrandmasterLeagueListDto
    _renamed = {"leagueId": "id"}

    def __call__(self, **kwargs):
        if "entries" in kwargs:
            self.entries = [LeagueEntryData(**entry) for entry in kwargs.pop("entries")]
        super().__call__(**kwargs)
        return self

class MasterLeagueListData(CoreData):
    _dto_type = MasterLeagueListDto
    _renamed = {"leagueId": "id"}

    def __call__(self, **kwargs):
        if "entries" in kwargs:
            self.entries = [LeagueEntryData(**entry) for entry in kwargs.pop("entries")]
        super().__call__(**kwargs)
        return self


##############
# Core Types #
##############


class MiniSeries(CassiopeiaObject):
    _data_types = {MiniSeriesData}
    # Technically wins, loses, and wins_required can all be calculated from progress, so we don't technically need to store those data.

    def __str__(self):
        return f"Promos{self.progress + [None for _ in range(self.not_played)]}"

    @property
    def wins(self) -> int:
        return self._data[MiniSeriesData].wins  # sum(self.progress)  # This will work too

    @property
    def losses(self) -> int:
        return self._data[MiniSeriesData].losses  # len(self._data[MiniSeriesData].progress[0]) - sum(self.progress)  # This will work too

    @property
    def wins_required(self) -> int:
        """2 or 3 wins will be required for promotion."""
        return self._data[MiniSeriesData].target  # {3: 2, 5: 3}[len(self._data[MiniSeriesData].progress[0])]

    @property
    def not_played(self) -> int:
        """The number of games in the player's promos that they haven't played yet."""
        return len(self._data[MiniSeriesData].progress) - len(self.progress)

    @lazy_property
    def progress(self) -> List[bool]:
        """A list of True/False for the number of games the played in the mini series indicating if the player won or lost."""
        return [True if p == "W" else False for p in self._data[MiniSeriesData].progress if p != "N"]


@searchable({str: ["division", "name", "summoner"], bool: ["hot_streak", "veteran", "fresh_blood"], Division: ["division"], Summoner: ["summoner"], Queue: ["queue"]})
class LeagueEntry(CassiopeiaGhost):
    _data_types = {LeagueEntryData}

    @provide_default_region
    def __init__(self, *, region: Union[Region, str] = None):
        kwargs = {"region": region}
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform}

    def __eq__(self, other: "LeagueEntry"):
        if not isinstance(other, LeagueEntry) or self.region != other.region:
            return False
        return self.summoner == other.summoner and self.queue == other.queue and self.name == other.name

    __hash__ = CassiopeiaGhost.__hash__

    @classmethod
    def from_data(cls, data: LeagueEntryData, loaded_groups: Optional[Set[Type[CoreData]]] = None, league: "League" = None):
        self = super().from_data(data=data, loaded_groups=loaded_groups)
        self.__league = league
        return self

    @lazy_property
    def region(self) -> Region:
        """The region for this champion."""
        return Region(self._data[LeagueEntryData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this champion."""
        return self.region.platform

    @lazy_property
    def queue(self) -> Queue:
        try:
            return Queue(self._data[LeagueEntryData].queue)
        except AttributeError:
            return self.league.queue

    @lazy_property
    def tier(self) -> Tier:
        return Tier(self._data[LeagueEntryData].tier)

    @lazy_property
    def division(self) -> Division:
        return Division(self._data[LeagueEntryData].division)

    @property
    def hot_streak(self) -> bool:
        return self._data[LeagueEntryData].hotStreak

    @lazy_property
    def promos(self) -> Optional[MiniSeries]:
        if hasattr(self._data[LeagueEntryData], "promos"):
            return MiniSeries.from_data(self._data[LeagueEntryData].promos)
        else:
            # Return None if the summoner isn't in their promos
            if hasattr(self._data[LeagueEntryData], "leaguePoints"):
                return None
        # Reraise the original error
        return MiniSeries.from_data(self._data[LeagueEntryData].promos)

    @property
    def wins(self) -> int:
        return self._data[LeagueEntryData].wins

    @property
    def veteran(self) -> bool:
        return self._data[LeagueEntryData].veteran

    @property
    def losses(self) -> int:
        return self._data[LeagueEntryData].losses

    @lazy_property
    def summoner(self) -> Summoner:
        return Summoner(id=self._data[LeagueEntryData].summonerId, name=self._data[LeagueEntryData].summonerName, region=self.region)

    @property
    def fresh_blood(self) -> bool:
        return self._data[LeagueEntryData].freshBlood

    @lazy_property
    def league(self) -> "League":
        return self.__league or League(id=self._data[LeagueEntryData].leagueId, region=self.region)

    @property
    def league_points(self) -> int:
        return self._data[LeagueEntryData].leaguePoints

    @property
    def inactive(self) -> bool:
        return self._data[LeagueEntryData].inactive

    @property
    def role(self) -> Role:
        return Role.from_league_naming_scheme(self._data[LeagueEntryData].role)


class LeagueEntries(CassiopeiaLazyList):  # type List[LeagueEntry]
    _data_types = {LeagueEntriesData}

    @provide_default_region
    def __init__(self, *, region: Union[Region, str] = None, queue: Queue = None, tier: Tier = None, division: Division = None):
        kwargs = {"region": region, "queue": queue, "tier": tier, "division": division}
        CassiopeiaObject.__init__(self, **kwargs)

    @classmethod
    @provide_default_region
    def __get_query_from_kwargs__(cls, *, region: Union[Region, str] = None, queue: Queue = None, tier: Tier = None, division: Division = None):
        query = {"region": region, "queue": queue, "tier": tier, "division": division}
        return query

    @classmethod
    def from_generator(cls, generator: Generator, region: Union[Region, str] = None, queue: Queue = None, tier: Tier = None, division: Division = None, **kwargs):
        self = cls.__new__(cls)
        kwargs["region"] = region
        kwargs["queue"] = queue
        kwargs["tier"] = tier
        kwargs["division"] = division
        CassiopeiaLazyList.__init__(self, generator=generator, **kwargs)
        return self

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[LeagueEntriesData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @lazy_property
    def queue(self) -> Queue:
        return Queue(self._data[LeagueEntriesData].queue)

    @lazy_property
    def tier(self) -> Tier:
        return Tier(self._data[LeagueEntriesData].tier)

    @lazy_property
    def division(self) -> Division:
        return Division(self._data[LeagueEntriesData].division)


class LeagueSummonerEntries(CassiopeiaLazyList):
    _data_types = {LeagueSummonerEntriesData}

    def __init__(self, *, summoner: Summoner):
        self.__summoner = summoner
        kwargs = {"region": summoner.region}
        CassiopeiaObject.__init__(self, **kwargs)

    @classmethod
    def __get_query_from_kwargs__(cls, *, summoner: Union[Summoner, str]) -> dict:
        query = {"region": summoner.region}
        if isinstance(summoner, Summoner):
            query["summoner.id"] = summoner.id
        elif isinstance(summoner, str):
            if len(summoner) < 35:
                query["summoner.id"] = Summoner(name=summoner, region=summoner.region).id
            else:
                query["summoner.id"] = summoner
        assert "summoner.id" in query
        return query

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[LeagueSummonerEntriesData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @lazy_property
    def fives(self):
        for entry in self:
            if entry.queue is Queue.ranked_solo_fives:
                return entry
        raise ValueError("Queue does not exist for this summoner.")

    @lazy_property
    def flex(self):
        for entry in self:
            if entry.queue is Queue.ranked_flex_fives:
                return entry
        raise ValueError("Queue does not exist for this summoner.")

    @lazy_property
    def threes(self):
        for entry in self:
            if entry.queue is Queue.ranked_flex_threes:
                return entry
        raise ValueError("Queue does not exist for this summoner.")


@searchable({str: ["tier", "queue", "name"], Queue: ["queue"], Tier: ["tier"]})
class League(CassiopeiaGhost):
    _data_types = {LeagueData}

    def __init__(self, id: str = None, queue: Queue = None, region: Union[Region, str] = None):
        if region is None:
            region = configuration.settings.default_region
        if region is not None and not isinstance(region, Region):
            region = Region(region)
        kwargs = {"id": id, "region": region, "queue": queue}
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"id": self.id, "region": self.region, "platform": self.platform}

    def __eq__(self, other: "League"):
        if not isinstance(other, League) or self.region != other.region:
            return False
        return self.id == other.id

    def __str__(self):
        region = self.region
        id_ = self.id
        return "League(id={id_}, region='{region}')".format(id_=id_, region=region.value)

    __hash__ = CassiopeiaGhost.__hash__

    def __getitem__(self, item):
        return self.entries[item]

    def __len__(self):
        return len(self.entries)

    def __iter__(self):
        return iter(self.entries)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[LeagueData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def id(self) -> str:
        return self._data[LeagueData].id

    @CassiopeiaGhost.property(LeagueData)
    @ghost_load_on
    @lazy
    def tier(self) -> Tier:
        return Tier(self._data[LeagueData].tier)

    @CassiopeiaGhost.property(LeagueData)
    @ghost_load_on
    @lazy
    def queue(self) -> Queue:
        return Queue(self._data[LeagueData].queue)

    @CassiopeiaGhost.property(LeagueData)
    @ghost_load_on
    def name(self) -> str:
        return self._data[LeagueData].name

    @CassiopeiaGhost.property(LeagueData)
    @ghost_load_on
    @lazy
    def entries(self) -> List[LeagueEntry]:
        entries = []
        for entry in self._data[LeagueData].entries:
            entry.leagueId = self.id
            entry = LeagueEntry.from_data(data=entry, loaded_groups={LeagueEntriesData})
            entries.append(entry)
        return SearchableList(entries)


class ChallengerLeague(League):
    _data_types = {ChallengerLeagueListData}

    @provide_default_region
    def __init__(self, *, queue: Union[Queue, str, int] = None, region: Union[Region, str] = None):
        kwargs = {"region": region}
        if isinstance(queue, int):
            kwargs["queue"] = Queue.from_id(queue)
        elif isinstance(queue, str):
            kwargs["queue"] = Queue(queue)
        elif isinstance(queue, Queue):
            kwargs["queue"] = queue
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "queue": self.queue}

    def __eq__(self, other: "ChallengerLeague"):
        if not isinstance(other, ChallengerLeague) or self.region != other.region:
            return False
        return self.queue == other.queue

    __hash__ = CassiopeiaGhost.__hash__

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[ChallengerLeagueListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform
    
    @CassiopeiaGhost.property(ChallengerLeagueListData)
    @ghost_load_on
    def id(self) -> str:
        return self._data[ChallengerLeagueListData].id

    @lazy_property
    def tier(self) -> Tier:
        return Tier.challenger

    @lazy_property
    def queue(self) -> Queue:
        return Queue(self._data[ChallengerLeagueListData].queue)

    @CassiopeiaGhost.property(ChallengerLeagueListData)
    @ghost_load_on
    def name(self) -> str:
        return self._data[ChallengerLeagueListData].name

    @CassiopeiaGhost.property(ChallengerLeagueListData)
    @ghost_load_on
    @lazy
    def entries(self) -> List[LeagueEntry]:
        return SearchableList([LeagueEntry.from_data(entry) for entry in self._data[ChallengerLeagueListData].entries])

class GrandmasterLeague(CassiopeiaGhost):
    _data_types = {GrandmasterLeagueListData}

    @provide_default_region
    def __init__(self, *, queue: Union[Queue, str, int] = None, region: Union[Region, str] = None):
        kwargs = {"region": region}
        if isinstance(queue, int):
            kwargs["queue"] = Queue.from_id(queue)
        elif isinstance(queue, str):
            kwargs["queue"] = Queue(queue)
        elif isinstance(queue, Queue):
            kwargs["queue"] = queue
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "queue": self.queue}

    def __eq__(self, other: "GrandmasterLeague"):
        if not isinstance(other, GrandmasterLeague) or self.region != other.region:
            return False
        return self.queue == other.queue

    __hash__ = CassiopeiaGhost.__hash__

    def __getitem__(self, item):
        return self.entries[item]

    def __len__(self):
        return len(self.entries)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[GrandmasterLeagueListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @CassiopeiaGhost.property(GrandmasterLeagueListData)
    @ghost_load_on
    def id(self) -> str:
        return self._data[GrandmasterLeagueListData].id

    @lazy_property
    def tier(self) -> Tier:
        return Tier.grandmaster

    @lazy_property
    def queue(self) -> Queue:
        return Queue(self._data[GrandmasterLeagueListData].queue)

    @CassiopeiaGhost.property(GrandmasterLeagueListData)
    @ghost_load_on
    def name(self) -> str:
        return self._data[GrandmasterLeagueListData].name

    @CassiopeiaGhost.property(GrandmasterLeagueListData)
    @ghost_load_on
    @lazy
    def entries(self) -> List[LeagueEntry]:
        return SearchableList([LeagueEntry.from_data(entry) for entry in self._data[GrandmasterLeagueListData].entries])

class MasterLeague(CassiopeiaGhost):
    _data_types = {MasterLeagueListData}

    @provide_default_region
    def __init__(self, *, queue: Union[Queue, str, int] = None, region: Union[Region, str] = None):
        kwargs = {"region": region}
        if isinstance(queue, int):
            kwargs["queue"] = Queue.from_id(queue)
        elif isinstance(queue, str):
            kwargs["queue"] = Queue(queue)
        elif isinstance(queue, Queue):
            kwargs["queue"] = queue
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "queue": self.queue}

    def __eq__(self, other: "MasterLeague"):
        if not isinstance(other, MasterLeague) or self.region != other.region:
            return False
        return self.queue == other.queue

    __hash__ = CassiopeiaGhost.__hash__

    def __getitem__(self, item):
        return self.entries[item]

    def __len__(self):
        return len(self.entries)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[MasterLeagueListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @CassiopeiaGhost.property(MasterLeagueListData)
    @ghost_load_on
    def id(self) -> str:
        return self._data[MasterLeagueListData].id

    @lazy_property
    def tier(self) -> Tier:
        return Tier.master

    @lazy_property
    def queue(self) -> Queue:
        return Queue(self._data[MasterLeagueListData].queue)

    @CassiopeiaGhost.property(MasterLeagueListData)
    @ghost_load_on
    def name(self) -> str:
        return self._data[MasterLeagueListData].name

    @CassiopeiaGhost.property(MasterLeagueListData)
    @ghost_load_on
    @lazy
    def entries(self) -> List[LeagueEntry]:
        return SearchableList([LeagueEntry.from_data(entry) for entry in self._data[MasterLeagueListData].entries])



import copy
from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator

from datapipelines import DataSource, PipelineContext, Query, NotFoundError, validate_query
from .common import KernelSource, APINotFoundError
from ...data import Platform
from ...dto.champion import ChampionRotationDto
from ..uniquekeys import convert_region_to_platform

T = TypeVar("T")


class ChampionAPI(KernelSource):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    _validate_get_champion_rotation_query = Query. \
        has("platform").as_(Platform)

    @get.register(ChampionRotationDto)
    @validate_query(_validate_get_champion_rotation_query, convert_region_to_platform)
    def get_champion_rotation(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionRotationDto:
        parameters = {"platform": query["platform"].value}
        endpoint = "lol/platform/v3/champion-rotations"
        try:
            data = self._get(endpoint=endpoint, parameters=parameters)
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["platform"] = query["platform"].value
        data["region"] = query["platform"].region.value
        return ChampionRotationDto(**data)

    _validate_get_many_champion_rotation_query = Query. \
        has("platforms").as_(Iterable)

    @get_many.register(ChampionRotationDto)
    @validate_query(_validate_get_many_champion_rotation_query, convert_region_to_platform)
    def get_many_champion_rotations(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ChampionRotationDto, None, None]:
        def generator():
            for platform in query["platforms"]:
                new_query = copy.deepcopy(query)
                new_query.pop("platforms")
                new_query["platform"] = platform
                yield self.get_champion_rotation(query=new_query)

        return generator()



import time
import copy
import functools
import collections
from abc import abstractmethod, ABC
from typing import MutableMapping, Any, Union, TypeVar, Iterable, Type, List, Tuple, Dict, Callable

from datapipelines import DataSource, PipelineContext
from merakicommons.ratelimits import FixedWindowRateLimiter, MultiRateLimiter

from ..common import HTTPClient, HTTPError, Curl
from ...data import Platform
from ...dto.staticdata.realm import RealmDto


def _get_latest_version(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    pipeline = context[PipelineContext.Keys.PIPELINE]
    realms = pipeline.get(RealmDto, {"platform": query["platform"]})
    return realms["v"]


def _get_default_locale(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    return query["platform"].default_locale


class APIRequestError(HTTPError):
    pass


class APIError(HTTPError):
    pass


class APINotFoundError(HTTPError):
    pass


class APIForbiddenError(APINotFoundError):
    pass


_ERROR_CODES = {
    400: APIRequestError,
    401: APIForbiddenError,
    403: APIRequestError,
    404: APINotFoundError,
    415: RuntimeError,
    429: RuntimeError,
    500: APIError,
    502: APIError,
    503: APIError,
    504: APIError
}

T = TypeVar("T")



class RiotAPIRateLimiter(MultiRateLimiter):
    # The application limiter and method limiters will each be an instance of this.

    def __init__(self, limiting_share):
        self.limiting_share = limiting_share
        super().__init__()  # Initialize with no underlying limiters
        self._limiters = []  # Make it a list rather than a tuple so we can append

    def restrict_for(self, seconds: int) -> None:
        for limiter in self._limiters:
            limiter.restrict_for(seconds)

    def _construct_limiters(self, limits: List[List[int]]):
        # Creates the necessary FixedWindowRateLimiters from the rates in the headers
        assert len(self._limiters) == 0
        # Create the rate limiters
        for permits, window in limits:
            self._limiters.append(
                FixedWindowRateLimiter(window_seconds=window, window_permits=permits)
            )

    def adjust_rate_limits_if_necessary(self, limits: List[List[int]]) -> None:
        if len(self._limiters) == 0:
            self._construct_limiters(limits)
        for permits, window in limits:
            permits = permits * self.limiting_share
            for_window = self._get_specific_limiter_for_window(window)
            if permits != for_window._window_permits:
                for_window.set_permits(permits)

    def _get_specific_limiter_for_window(self, window: int) -> FixedWindowRateLimiter:
        for limiter in self._limiters:
            if limiter._window_seconds == window:
                return limiter


def _split_rate_limit_header(header):
    rates = []
    for pw in header.split(","):
        pw = pw.split(":")
        p, w = pw
        p, w = int(p), int(w)
        rates.append((p, w))
    return rates


class RiotAPIService(DataSource):
    def __init__(self, api_key: str, app_rate_limiter: Dict[Platform, RiotAPIRateLimiter], request_error_handling: Dict = None, http_client: HTTPClient = None):
        self._limiting_share = app_rate_limiter[Platform.north_america].limiting_share

        if http_client is None:
            self._client = HTTPClient()
        else:
            self._client = http_client

        self._headers = {
            "X-Riot-Token": api_key
        }

        # Both the application and method rate limiters will be in the same rate limiter
        self._rate_limiters = {
            "application": app_rate_limiter
        }

        default_request_error_handling = {
            "404": {
                "strategy": "throw"
            },
            "429": {
                "service": {
                    "strategy": "exponential_backoff",
                    "initial_backoff": 1.0,
                    "backoff_factor": 2.0,
                    "max_attempts": 4
                },
                "method": {
                    "strategy": "retry_from_headers",
                    "max_attempts": 5
                },
                "application": {
                    "strategy": "retry_from_headers",
                    "max_attempts": 5
                }
            },
            "500": {
                "strategy": "throw"
            },
            "503": {
                "strategy": "throw"
            },
            "timeout": {
                "strategy": "throw"
            },
            "403": {
                "strategy": "throw"
            },
            "504": {
                "strategy": "exponential_backoff",
                "initial_backoff": 1.0,
                "backoff_factor": 2.0,
                "max_attempts": 4
            },
            "502": {
                "strategy": "exponential_backoff",
                "initial_backoff": 1.0,
                "backoff_factor": 2.0,
                "max_attempts": 4
            }
        }
        if request_error_handling is None:
            request_error_handling = default_request_error_handling
        else:
            def recursive_setdefault(d, u):
                for k, v in u.items():
                    if isinstance(v, collections.Mapping):
                        r = recursive_setdefault(d.get(k, {}), v)
                        d.setdefault(k, r)
                    else:
                        d.setdefault(k, u[k])
                return d
            recursive_setdefault(request_error_handling, default_request_error_handling)

        new_handler_instance = {
            "throw": lambda **init_args: ThrowException(),
            "exponential_backoff": lambda **init_args: ExponentialBackoff(**init_args),
            "retry_from_headers": lambda **init_args: RetryFromHeaders(**init_args)
        }
        self._handlers = {429: {}}  # type: Dict[Union[str, int], Union[Dict[Union[str, int], Callable], Callable]]
        for code, config in request_error_handling.items():
            config = copy.deepcopy(config)
            if code != "timeout":
                code = int(code)
            if code == 429:
                # config == {service: ..., method: ... , application: ...}
                for app, config in config.items():
                    strategy = config.pop("strategy")
                    self._handlers[code][app] = functools.partial(new_handler_instance[strategy], **config)
            else:
                strategy = config.pop("strategy")
                self._handlers[code] = functools.partial(new_handler_instance[strategy], **config)

    def _get_rate_limiter(self, platform: Platform, endpoint: str):
        try:
            method_limiter = self._rate_limiters[(platform, endpoint)]
        except KeyError:
            method_limiter = RiotAPIRateLimiter(self._limiting_share)
            self._rate_limiters[(platform, endpoint)] = method_limiter
        app_limiter = self._rate_limiters["application"][platform]
        return app_limiter, method_limiter

    def _adjust_rate_limiters_from_headers(self, app_limiter, method_limiter, response_headers):
        # If Riot changes the # of permits allowed in their response headers, change our rate limiters.
        # We are currently ignoring the X-*-Rate-Limit-Count headers and assuming our rate limiter logic agrees.
        if "X-App-Rate-Limit" in response_headers:
            limits = _split_rate_limit_header(response_headers["X-App-Rate-Limit"])
            app_limiter.adjust_rate_limits_if_necessary(limits)
        if "X-Method-Rate-Limit" in response_headers:
            limits = _split_rate_limit_header(response_headers["X-Method-Rate-Limit"])
            method_limiter.adjust_rate_limits_if_necessary(limits)

    def _get(self, url: str, parameters: MutableMapping[str, Any] = None, app_limiter: RiotAPIRateLimiter = None, method_limiter: RiotAPIRateLimiter = None, connection: Curl = None) -> Union[dict, list, Any]:
        # Make a new RiotAPIRequest and run it until it returns or fails.
        # If it returns, return the result.
        # If it fails, throw an appropriate error.
        request = RiotAPIRequest(service=self, url=url, parameters=parameters, app_limiter=app_limiter, method_limiter=method_limiter, connection=connection)
        try:
            return request()
        except HTTPError as error:
            # The error handlers didn't work, so raise an appropriate error.
            new_error_type = _ERROR_CODES[error.code]
            if new_error_type is RuntimeError:
                new_error = RuntimeError("Encountered an HTTP error code {code} with message \"{message}\" which should have already been handled. Report this to the Cassiopeia team.".format(code=error.code, message=str(error)))
            elif new_error_type is APIError:
                new_error = APIError("The Riot API experienced an internal error on the request. You may want to retry the request after a short wait or continue without the result. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            elif new_error_type is APINotFoundError:
                new_error = APINotFoundError("The Riot API returned a NOT FOUND error for the request. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            elif new_error_type is APIRequestError:
                new_error = APIRequestError("The Riot API returned an error on the request. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            elif new_error_type is APIForbiddenError:
                new_error = APIForbiddenError("The Riot API returned a FORBIDDEN error for the request. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            else:
                new_error = new_error_type(str(error))

            raise new_error from error

    @abstractmethod
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @abstractmethod
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass



class RiotAPIRequest(object):
    def __init__(self, service: RiotAPIService, url: str, parameters: MutableMapping[str, Any], app_limiter: RiotAPIRateLimiter, method_limiter: RiotAPIRateLimiter, connection: Curl):
        self.service = service
        self.url = url
        self.parameters = parameters
        self.app_limiter = app_limiter
        self.method_limiter = method_limiter
        self.connection = connection

    def __call__(self):
        try:
            body, response_headers = self.service._client.get(url=self.url,
                                                              parameters=self.parameters,
                                                              headers=self.service._headers,
                                                              rate_limiters=[self.app_limiter,
                                                                             self.method_limiter],
                                                              connection=self.connection)
            self.service._adjust_rate_limiters_from_headers(app_limiter=self.app_limiter,
                                                            method_limiter=self.method_limiter,
                                                            response_headers=response_headers)
            return body
        except HTTPError as error:
            return self._retry_request_by_handling_error(error)

    def _retry_request_by_handling_error(self, error: HTTPError, handlers=None):
            if handlers is None:
                handlers = []
            # Try to properly handling the 429 and retry the call after the appropriate time limit.
            if error.code == 429:
                # Identify which rate limit was hit (application, method, or service)
                if "X-Rate-Limit-Type" not in error.response_headers:
                    rate_limiting_type = "service"
                elif error.response_headers["X-Rate-Limit-Type"] == "application":
                    rate_limiting_type = "application"
                elif error.response_headers["X-Rate-Limit-Type"] == "method":
                    rate_limiting_type = "method"
                elif error.response_headers["X-Rate-Limit-Type"] == "service":
                    rate_limiting_type = "service"
                else:
                    raise ValueError("Unknown cause of rate limit; aborting. Headers were: {}".format(error.response_headers))

                # Create a new handler
                new_handler = self.service._handlers[429][rate_limiting_type]()  # type: FailedRequestHandler
            else:
                new_handler = self.service._handlers[error.code]()

            # If we will handle the new error in the same way as we did previously, don't use a new instance
            for handler in handlers:
                if isinstance(new_handler, handler.__class__):
                    new_handler = handler
                    break

            if new_handler.stop:
                raise error
            else:
                try:
                    body, response_headers = new_handler(error=error,
                                                         requester=self.service._client.get,
                                                         url=self.url,
                                                         parameters=self.parameters,
                                                         headers=self.service._headers,
                                                         rate_limiters=[self.app_limiter, self.method_limiter],
                                                         connection=self.connection)
                    self.service._adjust_rate_limiters_from_headers(app_limiter=self.app_limiter,
                                                                    method_limiter=self.method_limiter,
                                                                    response_headers=response_headers)
                    return body
                except HTTPError as error:
                    if new_handler not in handlers:
                        handlers.append(new_handler)
                    return self._retry_request_by_handling_error(error, handlers=handlers)


class FailedRequestHandler(ABC):
    @abstractmethod
    def __call__(self, error, requester, url, parameters, headers, rate_limiters, connection) -> Tuple[Union[dict, list, str, bytes], dict]:
        pass


class ExponentialBackoff(FailedRequestHandler):
    def __init__(self, initial_backoff: int, backoff_factor: int, max_attempts: int):
        self.backoff = initial_backoff
        self.factor = backoff_factor
        self.max_attempts = max_attempts
        self.attempts = 0
        self.stop = False

    def __call__(self, error, requester, url, parameters, headers, rate_limiters, connection) ->  Tuple[Union[dict, list, str, bytes], dict]:
        if self.attempts >= self.max_attempts:
            self.stop = True
            raise error
        print("INFO: Unexpected {} error ({}), backing off for {} seconds.".format(headers.get('X-Rate-Limit-Type', 'service'), error.code, self.backoff))
        time.sleep(self.backoff)
        self.backoff = self.backoff * self.factor
        self.attempts += 1
        return requester(url, parameters, headers, rate_limiters, connection)


class RetryFromHeaders(object):
    def __init__(self, max_attempts: int):
        self.max_attempts = int(max_attempts)
        self.attempts = 0
        self.stop = False

    def __call__(self, error, requester, url, parameters, headers, rate_limiters, connection) -> Tuple[Union[dict, list, str, bytes], dict]:
        if self.attempts >= self.max_attempts:
            self.stop = True
            raise error
        backoff = int(error.response_headers["Retry-After"])
        print("INFO: Unexpected {} rate limit, backing off for {} seconds (from headers).".format(headers.get('X-Rate-Limit-Type', 'service'), backoff))
        for rate_limiter in rate_limiters:
            rate_limiter.restrict_for(backoff)
        self.attempts += 1
        return requester(url, parameters, headers, rate_limiters, connection)


class ThrowException(FailedRequestHandler):
    def __init__(self):
        self.stop  = True

    def __call__(self, error, requester, url, parameters, headers, rate_limiters, connection) -> Tuple[Union[dict, list, str, bytes], dict]:
        raise error



from typing import Union

from merakicommons.cache import lazy_property

from ...data import Region, Platform
from ..common import CassiopeiaObject, CoreDataList, CassiopeiaLazyList, provide_default_region
from ...dto.staticdata.language import LanguagesDto


class LanguagesData(CoreDataList):
    _dto_type = LanguagesDto
    _renamed = {}


class Locales(CassiopeiaLazyList):
    _data_types = {LanguagesData}

    @provide_default_region
    def __init__(self, *, region: Union[Region, str] = None):
        CassiopeiaObject.__init__(self, region=region)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[LanguagesData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform



from django.shortcuts import render, HttpResponse
from django_cassiopeia import cassiopeia as cass
from time import sleep
import json


# Create your views here.
def test(request):
    return render(request, "test/test.html")

def test_request(request, n):
    context = {
        "n" : n,
    }
    sleep(n/5)
    try:
        kalturi = cass.Summoner(name="Kalturi", region="NA")
        senna = cass.Champion(id=235)
        rune = cass.Rune(id=8112)
        match = cass.Match(id=3481455783)
        if n < 20:
            print(match.creation)
        elif n < 40:
            print(kalturi.profile_icon.id)
        elif n < 60:
            print(kalturi.match_history_uri)
        elif n < 80:
            print(senna.ally_tips)
        elif n < 100:
            print(rune.name)
        else:
            history = cass.MatchHistory(summoner=kalturi, begin_index=0, end_index=100)
            print(match.id for match in history)
    except:
        raise RuntimeError("Failed at request "+str(n))
    return HttpResponse(json.dumps(context), content_type="application/json")


from time import time
from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator, Union
import arrow
import datetime
import math

from datapipelines import DataSource, PipelineContext, Query, NotFoundError, validate_query

from .common import KernelSource, APINotFoundError
from ...data import Platform, Season, Queue, SEASON_IDS, QUEUE_IDS
from ...dto.match import MatchDto, MatchListDto, TimelineDto
from ..uniquekeys import convert_region_to_platform

T = TypeVar("T")


def _get_current_time(query: MutableMapping[str, Any], context: PipelineContext = None) -> int:
    return int(time()) * 1000


class MatchAPI(KernelSource):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    _validate_get_match_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform)

    @get.register(MatchDto)
    @validate_query(_validate_get_match_query, convert_region_to_platform)
    def get_match(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MatchDto:
        parameters = {"platform": query["platform"].value}
        endpoint = "lol/match/v4/matches/{id}".format(id=query["id"])
        try:
            data = self._get(endpoint=endpoint, parameters=parameters)
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["gameId"] = query["id"]
        data["region"] = query["platform"].region.value
        for p in data["participantIdentities"]:
            aid = p.get("player", {}).get("currentAccountId", None)
            if aid == 0:
                p["player"]["bot"] = True
        return MatchDto(data)

    _validate_get_many_match_query = Query. \
        has("ids").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(MatchDto)
    @validate_query(_validate_get_many_match_query, convert_region_to_platform)
    def get_many_match(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[MatchDto, None, None]:
        def generator():
            parameters = {"platform": query["platform"].value}
            for id in query["ids"]:
                endpoint = "lol/match/v4/matches/{id}".format(id=id)
                try:
                    data = self._get(endpoint=endpoint, parameters=parameters)
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error
                
                for participant in data["participants"]:
                    participant.setdefault("runes", [])
                for p in data["participantIdentities"]:
                    aid = p.get("player", {}).get("currentAccountId", None)
                    if aid == 0:
                        p["player"]["bot"] = True

                data["gameId"] = id
                data["region"] = query["platform"].region.value
                yield MatchDto(data)

        return generator()

    _validate_get_match_list_query = Query. \
        has("accountId").as_(str).also. \
        has("platform").as_(Platform).also. \
        has("beginTime").as_(int).also. \
        can_have("endTime").as_(int).also. \
        has("beginIndex").as_(int).also. \
        has("maxNumberOfMatches").as_(float).also. \
        can_have("seasons").as_(Iterable).also. \
        can_have("champion.ids").as_(Iterable).also. \
        can_have("queues").as_(Iterable)

    @get.register(MatchListDto)
    @validate_query(_validate_get_match_list_query, convert_region_to_platform)
    def get_match_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MatchListDto:
        parameters = {"platform": query["platform"].value}

        riot_index_interval = 100
        riot_date_interval = datetime.timedelta(days=7)

        begin_time = query["beginTime"]  # type: arrow.Arrow
        end_time = query.get("endTime", arrow.now())  # type: arrow.Arrow
        if isinstance(begin_time, int):
            begin_time = arrow.get(begin_time / 1000)
        if isinstance(end_time, int):
            end_time = arrow.get(end_time / 1000)

        def determine_calling_method(begin_time, end_time) -> str:
            """Returns either "by_date" or "by_index"."""
            matches_per_date_interval = 10  # This is an assumption
            seconds_per_day = (60 * 60 * 24)
            riot_date_interval_in_days = riot_date_interval.total_seconds() / seconds_per_day  # in units of days
            npulls_by_date = (end_time - begin_time).total_seconds() / seconds_per_day / riot_date_interval_in_days
            npulls_by_index = (arrow.now() - begin_time).total_seconds() / seconds_per_day / riot_date_interval_in_days * matches_per_date_interval / riot_index_interval
            if math.ceil(npulls_by_date) < math.ceil(npulls_by_index):
                by = "by_date"
            else:
                by = "by_index"
            return by

        calling_method = determine_calling_method(begin_time, end_time)

        if calling_method == "by_date":
            parameters["beginTime"] = begin_time.timestamp * 1000
            if "endTime" in query:
                parameters["endTime"] = min((begin_time + riot_date_interval).timestamp * 1000, query["endTime"])
            else:
                parameters["endTime"] = (begin_time + riot_date_interval).timestamp * 1000
        else:
            parameters["beginIndex"] = query["beginIndex"]
            parameters["endIndex"] = query["beginIndex"] + min(riot_index_interval, query["maxNumberOfMatches"])
            parameters["endIndex"] = int(parameters["endIndex"])

        if "seasons" in query:
            seasons = {Season(season) for season in query["seasons"]}
            parameters["season"] = {SEASON_IDS[season] for season in seasons}
        else:
            seasons = set()

        if "champion.ids" in query:
            champions = query["champion.ids"]
            parameters["champion"] = champions
        else:
            champions = set()

        if "queues" in query:
            queues = {Queue(queue) for queue in query["queues"]}
            parameters["queue"] = {QUEUE_IDS[queue] for queue in queues}
        else:
            queues = set()

        endpoint = "lol/match/v4/matchlists/by-account/{accountId}".format(accountId=query["accountId"])
        try:
            data = self._get(endpoint=endpoint, parameters=parameters)
        except APINotFoundError:
            data = {"matches": []}

        data["accountId"] = query["accountId"]
        data["region"] = query["platform"].region.value
        data["season"] = seasons
        data["champion"] = champions
        data["queue"] = queues
        if calling_method == "by_index":
            data["beginIndex"] = parameters["beginIndex"]
            data["endIndex"] = parameters["endIndex"]
            data["maxNumberOfMatches"] = query["maxNumberOfMatches"]
        else:
            data["beginTime"] = parameters["beginTime"]
            data["endTime"] = parameters["endTime"]
        for match in data["matches"]:
            match["accountId"] = query["accountId"]
            match["region"] = Platform(match["platformId"]).region.value
        return MatchListDto(data)

    _validate_get_many_match_list_query = Query. \
        has("accountIds").as_(Iterable).also. \
        has("platform").as_(Platform).also. \
        can_have("beginTime").as_(int).also. \
        can_have("endTime").as_(int).also. \
        can_have("beginIndex").as_(int).also. \
        can_have("endIndex").as_(int).also. \
        can_have("seasons").as_(Iterable).also. \
        can_have("champion.ids").as_(Iterable).also. \
        can_have("queues").as_(Iterable)

    @get_many.register(MatchListDto)
    @validate_query(_validate_get_many_match_list_query, convert_region_to_platform)
    def get_many_match_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[MatchListDto, None, None]:
        parameters = {"platform": query["platform"].value}

        if "beginIndex" in query:
            parameters["beginIndex"] = query["beginIndex"]

        if "endIndex" in query:
            parameters["endIndex"] = query["endIndex"]

        if "seasons" in query:
            seasons = {Season(season) for season in query["seasons"]}
            parameters["season"] = {SEASON_IDS[season] for season in seasons}
        else:
            seasons = set()

        if "champion.ids" in query:
            parameters["champion"] = {query["champion.ids"]}

        if "queues" in query:
            queues = {Queue(queue) for queue in query["queues"]}
            parameters["queue"] = {QUEUE_IDS[queue] for queue in queues}
        else:
            queues = set()

        def generator():
            for id in query["accountIds"]:
                endpoint = "lol/match/v4/matchlists/by-account/{accountId}".format(accountId=id)
                try:
                    data = self._get(endpoint=endpoint, parameters=parameters)
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["accountId"] = id
                data["region"] = query["platform"].region.value
                if "beginIndex" in query:
                    data["beginIndex"] = query["beginIndex"]
                if "endIndex" in query:
                    data["endIndex"] = query["endIndex"]
                if "seasons" in query:
                    data["seasons"] = seasons
                if "champion.ids" in query:
                    data["champion"] = parameters["champion"]
                if "queues" in query:
                    parameters["queue"] = queues
                yield MatchListDto(data)

        return generator()

    _validate_get_timeline_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform)

    @get.register(TimelineDto)
    @validate_query(_validate_get_timeline_query, convert_region_to_platform)
    def get_match_timeline(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> TimelineDto:
        parameters = {"platform": query["platform"].value}
        endpoint = "lol/match/v4/timelines/by-match/{id}".format(id=query["id"])
        try:
            data = self._get(endpoint=endpoint, parameters=parameters)
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["matchId"] = query["id"]
        data["region"] = query["platform"].region.value
        return TimelineDto(data)

    _validate_get_many_timeline_query = Query. \
        has("ids").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(TimelineDto)
    @validate_query(_validate_get_many_timeline_query, convert_region_to_platform)
    def get_many_match_timeline(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[TimelineDto, None, None]:
        parameters = {"platform": query["platform"].value}
        def generator():
            for id in query["ids"]:
                endpoint = "lol/match/v4/timelines/by-match/{id}".format(id=id)
                try:
                    data = self._get(endpoint=endpoint, parameters=parameters)
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["matchId"] = id
                data["region"] = query["platform"].region.value
                yield TimelineDto(data)

        return generator()



import arrow
import datetime
from typing import Union

from datapipelines import NotFoundError
from merakicommons.cache import lazy_property
from merakicommons.container import searchable

from ..data import Region, Platform, Rank
from .common import CoreData, CassiopeiaObject, CassiopeiaGhost, provide_default_region, ghost_load_on
from .staticdata import ProfileIcon
from ..dto.summoner import SummonerDto


##############
# Data Types #
##############


class SummonerData(CoreData):
    _dto_type = SummonerDto
    _renamed = {"summonerLevel": "level"}


##############
# Core Types #
##############


@searchable({str: ["name", "region", "platform", "id", "account_id", "puuid"], Region: ["region"], Platform: ["platform"]})
class Summoner(CassiopeiaGhost):
    _data_types = {SummonerData}

    @provide_default_region
    def __init__(self, *, id: str = None, account_id: str = None, puuid: str = None, name: str = None, region: Union[Region, str] = None):
        kwargs = {"region": region}

        if id is not None:
            kwargs["id"] = id
        if account_id is not None:
            kwargs["accountId"] = account_id
        if puuid is not None:
            kwargs["puuid"] = puuid
        if name is not None:
            kwargs["name"] = name
        super().__init__(**kwargs)

    @classmethod
    @provide_default_region
    def __get_query_from_kwargs__(cls, *, id: str = None, account_id: str=None, puuid: str=None, name: str = None, region: Union[Region, str]) -> dict:
        query = {"region": region}
        if id is not None:
            query["id"] = id
        if account_id is not None:
            query["accountId"] = account_id
        if puuid is not None:
            query["puuid"] = puuid
        if name is not None:
            query["name"] = name
        return query

    def __get_query__(self):
        query = {"region": self.region, "platform": self.platform}
        try:
            query["puuid"] = self._data[SummonerData].puuid
        except AttributeError:
            pass
        try:
            query["id"] = self._data[SummonerData].id
        except AttributeError:
            pass
        try:
            query["accountId"] = self._data[SummonerData].accountId
        except AttributeError:
            pass
        try:
            query["name"] = self._data[SummonerData].name
        except AttributeError:
            pass
        assert "id" in query or "name" in query or "accountId" in query or "puuid" in query
        return query

    def __eq__(self, other: "Summoner"):
        if not isinstance(other, Summoner) or self.region != other.region:
            return False
        s = {}
        o = {}
        if hasattr(self._data[SummonerData], "id"): s["id"] = self.id
        if hasattr(other._data[SummonerData], "id"): o["id"] = other.id
        if hasattr(self._data[SummonerData], "name"): s["name"] = self.sanitized_name
        if hasattr(other._data[SummonerData], "name"): o["name"] = other.sanitized_name
        if hasattr(self._data[SummonerData], "accountId"): s["accountId"] = self.account_id
        if hasattr(other._data[SummonerData], "accountId"): o["accountId"] = other.account_id
        if any(s.get(key, "s") == o.get(key, "o") for key in s):
            return True
        else:
            return self.id == other.id

    def __str__(self):
        id_ = "?"
        name = "?"
        if hasattr(self._data[SummonerData], "id"):
            id_ = self.id
        if hasattr(self._data[SummonerData], "name"):
            name = self.name
        try:
            account_id = self._data[SummonerData].accountId
        except AttributeError:
            account_id = "?"
        try:
            puuid = self._data[SummonerData].puuid
        except AttributeError:
            puuid = "?"
        return "Summoner(id={id_}, account_id={account_id}, name='{name}', puuid='{puuid}')".format(id_=id_, name=name, account_id=account_id, puuid=puuid)

    @property
    def exists(self):
        try:
            if not self._Ghost__all_loaded:
                self.__load__()
            self.revision_date  # Make sure we can access this attribute
            return True
        except (AttributeError, NotFoundError):
            return False

    @lazy_property
    def region(self) -> Region:
        """The region for this summoner."""
        return Region(self._data[SummonerData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this summoner."""
        return self.region.platform

    @CassiopeiaGhost.property(SummonerData)
    @ghost_load_on
    def account_id(self) -> str:
        return self._data[SummonerData].accountId

    @CassiopeiaGhost.property(SummonerData)
    @ghost_load_on
    def puuid(self) -> str:
        return self._data[SummonerData].puuid

    @CassiopeiaGhost.property(SummonerData)
    @ghost_load_on
    def id(self) -> str:
        return self._data[SummonerData].id

    @CassiopeiaGhost.property(SummonerData)
    @ghost_load_on
    def name(self) -> str:
        return self._data[SummonerData].name

    @property
    def sanitized_name(self) -> str:
        return self.name.replace(" ", "").lower()

    @CassiopeiaGhost.property(SummonerData)
    @ghost_load_on
    def level(self) -> str:
        return self._data[SummonerData].level

    @CassiopeiaGhost.property(SummonerData)
    @ghost_load_on
    def profile_icon(self) -> ProfileIcon:
        return ProfileIcon(id=self._data[SummonerData].profileIconId, region=self.region)

    @CassiopeiaGhost.property(SummonerData)
    @ghost_load_on
    def revision_date(self) -> datetime.datetime:
        return arrow.get(self._data[SummonerData].revisionDate / 1000)

    @property
    def match_history_uri(self) -> str:
        return self.match_history[0].participants[self].match_history_uri

    # Special core methods

    @property
    def champion_masteries(self) -> "ChampionMasteries":
        from .championmastery import ChampionMasteries
        return ChampionMasteries(summoner=self, region=self.region)

    @property
    def match_history(self) -> "MatchHistory":
        from .match import MatchHistory
        return MatchHistory(summoner=self)

    @property
    def current_match(self) -> "CurrentMatch":
        from .spectator import CurrentMatch
        return CurrentMatch(summoner=self, region=self.region)

    @property
    def league_entries(self) -> "SummonerLeagues":
        from .league import LeagueSummonerEntries
        leagues = LeagueSummonerEntries(summoner=self)
        return leagues

    @lazy_property
    def rank_last_season(self):
        most_recent_match = self.match_history[0]
        return most_recent_match.participants[self.name].rank_last_season

    @property
    def verification_string(self) -> str:
        from .thirdpartycode import VerificationString
        vs = VerificationString(summoner=self, region=self.region)
        return vs.string

    @lazy_property
    def ranks(self):
        ranks = {}
        for position in self.league_entries:
            ranks[position.queue] = Rank(tier=position.tier, division=position.division)
        return ranks



from typing import Type, TypeVar, MutableMapping, Any, Iterable

from datapipelines import DataSource, PipelineContext, Query, NotFoundError, validate_query
from .common import KernelSource, APINotFoundError
from ...data import Platform, Region
from ...dto.staticdata.version import VersionListDto
from ...dto.spectator import CurrentGameInfoDto, FeaturedGamesDto
from ..uniquekeys import convert_region_to_platform

T = TypeVar("T")


def _get_default_version(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    pipeline = context[PipelineContext.Keys.PIPELINE]
    versions = pipeline.get(VersionListDto, {"platform": query["platform"]})
    return versions["versions"][0]


def _get_default_locale(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    return query["platform"].default_locale


class SpectatorAPI(KernelSource):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    ################
    # Current Game #
    ################

    _validate_get_current_game_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(str)

    @get.register(CurrentGameInfoDto)
    @validate_query(_validate_get_current_game_query, convert_region_to_platform)
    def get_current_game(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> CurrentGameInfoDto:
        parameters = {"platform": query["platform"].value}
        endpoint = "lol/spectator/v4/active-games/by-summoner/{id}".format(id=query["summoner.id"])
        try:
            data = self._get(endpoint=endpoint, parameters=parameters)
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["summonerId"] = query["summoner.id"]
        return CurrentGameInfoDto(data)

    #################
    # Featured Game #
    #################

    _validate_get_featured_game_query = Query. \
        has("platform").as_(Platform)

    @get.register(FeaturedGamesDto)
    @validate_query(_validate_get_featured_game_query, convert_region_to_platform)
    def get_featured_games(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> FeaturedGamesDto:
        parameters = {"platform": query["platform"].value}
        endpoint = "lol/spectator/v4/featured-games"
        try:
            data = self._get(endpoint=endpoint, parameters=parameters)
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        for game in data["gameList"]:
            game["region"] = data["region"]
        return FeaturedGamesDto(data)



from typing import Type, Mapping, Any, Iterable, TypeVar, Tuple, Callable, Generator
import datetime

from datapipelines import DataSource, DataSink, PipelineContext, validate_query, NotFoundError
from merakicommons.cache import Cache as CommonsCache

from . import uniquekeys
from ..core.staticdata.champion import ChampionData, ChampionListData, Champion, Champions
from ..core.staticdata.rune import RuneData, RuneListData, Rune, Runes
from ..core.staticdata.item import ItemData, ItemListData, Item, Items
from ..core.staticdata.summonerspell import SummonerSpellData, SummonerSpellListData, SummonerSpell, SummonerSpells
from ..core.staticdata.map import MapData, MapListData, Map, Maps
from ..core.staticdata.realm import RealmData, Realms
from ..core.staticdata.profileicon import ProfileIconData, ProfileIconListData, ProfileIcon, ProfileIcons
from ..core.staticdata.language import LanguagesData, Locales
from ..core.staticdata.languagestrings import LanguageStringsData, LanguageStrings
from ..core.staticdata.version import VersionListData, Versions
from ..core.championmastery import ChampionMasteryData, ChampionMasteryListData, ChampionMastery, ChampionMasteries
from ..core.league import MasterLeagueListData, GrandmasterLeagueListData, ChallengerLeagueListData, LeagueSummonerEntries, League, ChallengerLeague, GrandmasterLeague, MasterLeague, LeagueEntries
from ..core.match import MatchData, TimelineData, Match, Timeline
from ..core.summoner import SummonerData, Summoner
from ..core.status import ShardStatusData, ShardStatus
from ..core.spectator import CurrentGameInfoData, FeaturedGamesData, CurrentMatch, FeaturedMatches
from ..core.champion import ChampionRotationData, ChampionRotation

T = TypeVar("T")


default_expirations = {
    ChampionRotationData: datetime.timedelta(hours=6),
    Realms: datetime.timedelta(hours=6),
    Versions: datetime.timedelta(hours=6),
    Champion: datetime.timedelta(days=20),
    Rune: datetime.timedelta(days=20),
    Item: datetime.timedelta(days=20),
    SummonerSpell: datetime.timedelta(days=20),
    Map: datetime.timedelta(days=20),
    ProfileIcon: datetime.timedelta(days=20),
    Locales: datetime.timedelta(days=20),
    LanguageStrings: datetime.timedelta(days=20),
    SummonerSpells: datetime.timedelta(days=20),
    Items: datetime.timedelta(days=20),
    Champions: datetime.timedelta(days=20),
    Runes: datetime.timedelta(days=20),
    Maps: datetime.timedelta(days=20),
    ProfileIcons: datetime.timedelta(days=20),
    ChampionMastery: datetime.timedelta(days=7),
    ChampionMasteries: datetime.timedelta(days=7),
    LeagueSummonerEntries: datetime.timedelta(hours=6),
    League: datetime.timedelta(hours=6),
    ChallengerLeague: datetime.timedelta(hours=6),
    GrandmasterLeague: datetime.timedelta(hours=6),
    MasterLeague: datetime.timedelta(hours=6),
    Match: datetime.timedelta(days=3),
    Timeline: datetime.timedelta(days=1),
    Summoner: datetime.timedelta(days=1),
    ShardStatus: datetime.timedelta(hours=1),
    CurrentMatch: datetime.timedelta(hours=0.5),
    FeaturedMatches: datetime.timedelta(hours=0.5),
}


class Cache(DataSource, DataSink):
    def __init__(self, expirations: Mapping[type, float] = None) -> None:
        self._cache = CommonsCache()
        self._expirations = dict(expirations) if expirations is not None else default_expirations
        for key, value in list(self._expirations.items()):
            if isinstance(key, str):
                new_key = globals()[key]
                self._expirations[new_key] = self._expirations.pop(key)
                key = new_key
            if value != -1 and isinstance(value, datetime.timedelta):
                self._expirations[key] = value.seconds + 24 * 60 * 60 * value.days

    @DataSource.dispatch
    def get(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    @DataSink.dispatch
    def put(self, type: Type[T], item: T, context: PipelineContext = None) -> None:
        pass

    @DataSink.dispatch
    def put_many(self, type: Type[T], items: Iterable[T], context: PipelineContext = None) -> None:
        pass

    def _get(self, type: Type[T], query: Mapping[str, Any], key_function: Callable[[Mapping[str, Any]], Any], context: PipelineContext = None) -> T:
        keys = key_function(query)
        for key in keys:
            try:
                return self._cache.get(type, key)
            except KeyError:
                pass
        else:
            raise NotFoundError

    def _get_many(self, type: Type[T], query: Mapping[str, Any], key_generator: Callable[[Mapping[str, Any]], Any], context: PipelineContext = None) -> Generator[T, None, None]:
        for keys in key_generator(query):
            for key in keys:
                try:
                    yield self._cache.get(type, key)
                except KeyError:
                    pass
            else:
                raise NotFoundError

    @staticmethod
    def _put_many_generator(items: Iterable[T], key_function: Callable[[T], Any]) -> Generator[Tuple[Any, T], None, None]:
        for item in items:
            for key in key_function(item):
                yield key, item

    def _put(self, type: Type[T], item: T, key_function: Callable[[T], Any], context: PipelineContext = None) -> None:
        try:
            expire_seconds = self._expirations[type]
        except KeyError:
            expire_seconds = -1

        if expire_seconds != 0:
            keys = key_function(item)
            for key in keys:
                self._cache.put(type, key, item, expire_seconds)

    def _put_many(self, type: Type[T], items: Iterable[T], key_function: Callable[[T], Any], context: PipelineContext = None) -> None:
        expire_seconds = self._expirations.get(type, default_expirations[type])
        for key, item in Cache._put_many_generator(items, key_function):
            self._cache.put(type, key, item, expire_seconds)

    def clear(self, type: Type[T] = None):
        if type is None:
            for key in self._cache._data:
                self._cache._data[key].clear()
        else:
            self._cache._data[type].clear()

    def expire(self, type: Type[T] = None):
        self._cache.expire(type)


    #####################
    # Champion Rotation #
    #####################

    # Champion Rotation Data

    @get.register(ChampionRotationData)
    @validate_query(uniquekeys.validate_champion_rotation_query, uniquekeys.convert_region_to_platform)
    def get_champion_rotation(self, query: Mapping[str, Any], context: PipelineContext = None) -> ChampionRotationData:
        return self._get(ChampionRotationData, query, uniquekeys.for_champion_rotation_query, context)

    @get_many.register(ChampionRotationData)
    @validate_query(uniquekeys.validate_many_champion_rotation_query, uniquekeys.convert_region_to_platform)
    def get_many_champion_rotation(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[ChampionRotationData, None, None]:
        return self._get_many(ChampionRotationData, query, uniquekeys.for_many_champion_rotation_query, context)

    @put.register(ChampionRotationData)
    def put_champion_rotation(self, item: ChampionRotationData, context: PipelineContext = None) -> None:
        self._put(ChampionRotationData, item, uniquekeys.for_champion_rotation, context=context)

    @put_many.register(ChampionRotationData)
    def put_many_champion_rotation(self, items: Iterable[ChampionRotationData], context: PipelineContext = None) -> None:
        self._put_many(ChampionRotationData, items, uniquekeys.for_many_champion_rotation, context=context)


    ########################
    # Champion Mastery API #
    ########################

    # Champion Mastery

    @get.register(ChampionMastery)
    @validate_query(uniquekeys.validate_champion_mastery_query, uniquekeys.convert_region_to_platform)
    def get_champion_mastery(self, query: Mapping[str, Any], context: PipelineContext = None) -> ChampionMastery:
        return self._get(ChampionMastery, query, uniquekeys.for_champion_mastery_query, context)

    @get_many.register(ChampionMastery)
    @validate_query(uniquekeys.validate_many_champion_mastery_query, uniquekeys.convert_region_to_platform)
    def get_many_champion_mastery(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[ChampionMastery, None, None]:
        return self._get_many(ChampionMastery, query, uniquekeys.for_many_champion_mastery_query, context)

    @put.register(ChampionMastery)
    def put_champion_mastery(self, item: ChampionMastery, context: PipelineContext = None) -> None:
        self._put(ChampionMastery, item, uniquekeys.for_champion_mastery, context=context)

    @put_many.register(ChampionMastery)
    def put_many_champion_mastery(self, items: Iterable[ChampionMastery], context: PipelineContext = None) -> None:
        self._put_many(ChampionMastery, items, uniquekeys.for_many_champion_mastery, context=context)

    @get.register(ChampionMasteryData)
    @validate_query(uniquekeys.validate_champion_mastery_query, uniquekeys.convert_region_to_platform)
    def get_champion_mastery_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> ChampionMasteryData:
        result = self.get_champion_mastery(query=query, context=context)
        if result._data[ChampionMasteryData] is not None and result._Ghost__is_loaded(ChampionMasteryData):
            return result._data[ChampionMasteryData]
        else:
            raise NotFoundError

    # Champion Masteries

    @get.register(ChampionMasteries)
    @validate_query(uniquekeys.validate_champion_masteries_query, uniquekeys.convert_region_to_platform)
    def get_champion_masteries(self, query: Mapping[str, Any], context: PipelineContext = None) -> ChampionMasteries:
        return self._get(ChampionMasteries, query, uniquekeys.for_champion_masteries_query, context)

    @get_many.register(ChampionMasteries)
    @validate_query(uniquekeys.validate_many_champion_masteries_query, uniquekeys.convert_region_to_platform)
    def get_many_champion_masteries(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[ChampionMasteries, None, None]:
        return self._get_many(ChampionMasteries, query, uniquekeys.for_many_champion_masteries_query, context)

    @put.register(ChampionMasteries)
    def put_champion_masteries(self, item: ChampionMasteries, context: PipelineContext = None) -> None:
        self._put(ChampionMasteries, item, uniquekeys.for_champion_masteries, context=context)
        for cm in item:
            self._put(ChampionMastery, cm, uniquekeys.for_champion_mastery, context=context)

    @put_many.register(ChampionMasteries)
    def put_many_champion_masteries(self, items: Iterable[ChampionMasteries], context: PipelineContext = None) -> None:
        self._put_many(ChampionMasteries, items, uniquekeys.for_champion_masteries, context=context)

    ##############
    # League API #
    ##############

    # LeagueSummonerEntries

    @get.register(LeagueSummonerEntries)
    @validate_query(uniquekeys.validate_league_entries_query, uniquekeys.convert_region_to_platform)
    def get_league_summoner_entries(self, query: Mapping[str, Any], context: PipelineContext = None) -> LeagueSummonerEntries:
        return self._get(LeagueSummonerEntries, query, uniquekeys.for_league_summoner_entries_query, context)

    @get_many.register(LeagueSummonerEntries)
    @validate_query(uniquekeys.validate_many_league_entries_query, uniquekeys.convert_region_to_platform)
    def get_many_league_summoner_entries(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[LeagueSummonerEntries, None, None]:
        return self._get_many(LeagueSummonerEntries, query, uniquekeys.for_many_league_summoner_entries_query, context)

    @put.register(LeagueSummonerEntries)
    def put_league_summoner_entries(self, item: LeagueSummonerEntries, context: PipelineContext = None) -> None:
        self._put(LeagueSummonerEntries, item, uniquekeys.for_league_summoner_entries, context=context)

    @put_many.register(LeagueSummonerEntries)
    def put_many_league_summoner_entries(self, items: Iterable[LeagueSummonerEntries], context: PipelineContext = None) -> None:
        self._put_many(LeagueSummonerEntries, items, uniquekeys.for_league_summoner_entries, context=context)

    # Challenger

    @get.register(ChallengerLeague)
    @validate_query(uniquekeys.validate_challenger_league_query, uniquekeys.convert_region_to_platform)
    def get_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> ChallengerLeague:
        return self._get(ChallengerLeague, query, uniquekeys.for_challenger_league_query, context)

    @get_many.register(ChallengerLeague)
    @validate_query(uniquekeys.validate_many_challenger_league_query, uniquekeys.convert_region_to_platform)
    def get_many_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[ChallengerLeague, None, None]:
        return self._get_many(ChallengerLeague, query, uniquekeys.for_many_challenger_league_query, context)

    @put.register(ChallengerLeague)
    def put_league_summoner(self, item: ChallengerLeague, context: PipelineContext = None) -> None:
        self._put(ChallengerLeague, item, uniquekeys.for_challenger_league, context=context)

    @put_many.register(ChallengerLeague)
    def put_many_league_summoner(self, items: Iterable[ChallengerLeague], context: PipelineContext = None) -> None:
        self._put_many(ChallengerLeague, items, uniquekeys.for_challenger_league, context=context)

    # Grandmaster

    @get.register(GrandmasterLeague)
    @validate_query(uniquekeys.validate_grandmaster_league_query, uniquekeys.convert_region_to_platform)
    def get_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> GrandmasterLeague:
        return self._get(GrandmasterLeague, query, uniquekeys.for_grandmaster_league_query, context)

    @get_many.register(GrandmasterLeague)
    @validate_query(uniquekeys.validate_many_grandmaster_league_query, uniquekeys.convert_region_to_platform)
    def get_many_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[GrandmasterLeague, None, None]:
        return self._get_many(GrandmasterLeague, query, uniquekeys.for_many_grandmaster_league_query, context)

    @put.register(GrandmasterLeague)
    def put_league_summoner(self, item: GrandmasterLeague, context: PipelineContext = None) -> None:
        self._put(GrandmasterLeague, item, uniquekeys.for_grandmaster_league, context=context)

    @put_many.register(GrandmasterLeague)
    def put_many_league_summoner(self, items: Iterable[GrandmasterLeague], context: PipelineContext = None) -> None:
        self._put_many(GrandmasterLeague, items, uniquekeys.for_grandmaster_league, context=context)

    # Master

    @get.register(MasterLeague)
    @validate_query(uniquekeys.validate_master_league_query, uniquekeys.convert_region_to_platform)
    def get_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> MasterLeague:
        return self._get(MasterLeague, query, uniquekeys.for_master_league_query, context)

    @get_many.register(MasterLeague)
    @validate_query(uniquekeys.validate_many_master_league_query, uniquekeys.convert_region_to_platform)
    def get_many_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[MasterLeague, None, None]:
        return self._get_many(MasterLeague, query, uniquekeys.for_many_master_league_query, context)

    @put.register(MasterLeague)
    def put_league_summoner(self, item: MasterLeague, context: PipelineContext = None) -> None:
        self._put(MasterLeague, item, uniquekeys.for_master_league, context=context)

    @put_many.register(MasterLeague)
    def put_many_league_summoner(self, items: Iterable[MasterLeague], context: PipelineContext = None) -> None:
        self._put_many(MasterLeague, items, uniquekeys.for_master_league, context=context)


    ###################
    # Static Data API #
    ###################

    # Champion

    @get.register(Champion)
    @validate_query(uniquekeys.validate_champion_query, uniquekeys.convert_region_to_platform)
    def get_champion(self, query: Mapping[str, Any], context: PipelineContext = None) -> Champion:
        return self._get(Champion, query, uniquekeys.for_champion_query, context)

    @get_many.register(Champion)
    @validate_query(uniquekeys.validate_many_champion_query, uniquekeys.convert_region_to_platform)
    def get_many_champion(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Champion, None, None]:
        return self._get_many(Champion, query, uniquekeys.for_many_champion_query, context)

    @put.register(Champion)
    def put_champion(self, item: Champion, context: PipelineContext = None) -> None:
        self._put(Champion, item, uniquekeys.for_champion, context=context)

    @put_many.register(Champion)
    def put_many_champion(self, items: Iterable[Champion], context: PipelineContext = None) -> None:
        self._put_many(Champion, items, uniquekeys.for_champion, context=context)

    @get.register(ChampionData)
    @validate_query(uniquekeys.validate_champion_query, uniquekeys.convert_region_to_platform)
    def get_champion_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> ChampionData:
        result = self.get_champion(query=query, context=context)
        if result._data[ChampionData] is not None and result._Ghost__is_loaded(ChampionData):
            return result._data[ChampionData]
        else:
            raise NotFoundError

    # Champions

    @get.register(Champions)
    @validate_query(uniquekeys.validate_champions_query, uniquekeys.convert_region_to_platform)
    def get_champions(self, query: Mapping[str, Any], context: PipelineContext = None) -> Champions:
        return self._get(Champions, query, uniquekeys.for_champions_query, context)

    @get_many.register(Champions)
    @validate_query(uniquekeys.validate_many_champions_query, uniquekeys.convert_region_to_platform)
    def get_many_champions(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Champions, None, None]:
        return self._get_many(Champions, query, uniquekeys.for_many_champions_query, context)

    @put.register(Champions)
    def put_champions(self, champions: Champions, context: PipelineContext = None) -> None:
        self._put(Champions, champions, uniquekeys.for_champions, context=context)
        for champion in champions:
            self._put(Champion, champion, uniquekeys.for_champion, context=context)

    @put_many.register(Champions)
    def put_many_champions(self, champions: Iterable[Champions], context: PipelineContext = None) -> None:
        self._put_many(Champions, champions, uniquekeys.for_champions, context=context)

    # Item

    @get.register(Item)
    @validate_query(uniquekeys.validate_item_query, uniquekeys.convert_region_to_platform)
    def get_item(self, query: Mapping[str, Any], context: PipelineContext = None) -> Item:
        return self._get(Item, query, uniquekeys.for_item_query, context)

    @get_many.register(Item)
    @validate_query(uniquekeys.validate_many_item_query, uniquekeys.convert_region_to_platform)
    def get_many_item(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Item, None, None]:
        return self._get_many(Item, query, uniquekeys.for_many_item_query, context)

    @put.register(Item)
    def put_item(self, item: Item, context: PipelineContext = None) -> None:
        self._put(Item, item, uniquekeys.for_item, context=context)

    @put_many.register(Item)
    def put_many_item(self, items: Iterable[Item], context: PipelineContext = None) -> None:
        self._put_many(Item, items, uniquekeys.for_item, context=context)

    @get.register(ItemData)
    @validate_query(uniquekeys.validate_item_query, uniquekeys.convert_region_to_platform)
    def get_item_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> ItemData:
        result = self.get_item(query=query, context=context)
        if result._data[ItemData] is not None and result._Ghost__is_loaded(ItemData):
            return result._data[ItemData]
        else:
            raise NotFoundError

    # Items

    @get.register(Items)
    @validate_query(uniquekeys.validate_items_query, uniquekeys.convert_region_to_platform)
    def get_items(self, query: Mapping[str, Any], context: PipelineContext = None) -> Items:
        return self._get(Items, query, uniquekeys.for_items_query, context=context)

    @get_many.register(Items)
    @validate_query(uniquekeys.validate_many_items_query, uniquekeys.convert_region_to_platform)
    def get_many_items(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Items, None, None]:
        return self._get_many(Items, query, uniquekeys.for_many_items_query, context)

    @put.register(Items)
    def put_items(self, items: Items, context: PipelineContext = None) -> None:
        self._put(Items, items, uniquekeys.for_items, context=context)
        for item in items:
            self._put(Item, item, uniquekeys.for_item, context=context)

    @put_many.register(Items)
    def put_many_items(self, many_items: Iterable[Items], context: PipelineContext = None) -> None:
        self._put_many(Items, many_items, uniquekeys.for_items, context=context)

    # Language

    @get.register(Locales)
    @validate_query(uniquekeys.validate_languages_query, uniquekeys.convert_region_to_platform)
    def get_languages(self, query: Mapping[str, Any], context: PipelineContext = None) -> Locales:
        return self._get(Locales, query, uniquekeys.for_languages_query, context)

    @get_many.register(Locales)
    @validate_query(uniquekeys.validate_many_languages_query, uniquekeys.convert_region_to_platform)
    def get_many_languages(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Locales, None, None]:
        return self._get_many(Locales, query, uniquekeys.for_many_languages_query, context)

    @put.register(Locales)
    def put_languages(self, item: Locales, context: PipelineContext = None) -> None:
        self._put(Locales, item, uniquekeys.for_languages, context=context)

    @put_many.register(Locales)
    def put_many_languages(self, items: Iterable[Locales], context: PipelineContext = None) -> None:
        self._put_many(Locales, items, uniquekeys.for_languages, context=context)

    # Language strings

    @get.register(LanguageStrings)
    @validate_query(uniquekeys.validate_language_strings_query, uniquekeys.convert_region_to_platform)
    def get_language_strings(self, query: Mapping[str, Any], context: PipelineContext = None) -> LanguageStrings:
        return self._get(LanguageStrings, query, uniquekeys.for_language_strings_query, context)

    @get_many.register(LanguageStrings)
    @validate_query(uniquekeys.validate_many_language_strings_query, uniquekeys.convert_region_to_platform)
    def get_many_language_strings(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[LanguageStrings, None, None]:
        return self._get_many(LanguageStrings, query, uniquekeys.for_many_language_strings_query, context)

    @put.register(LanguageStrings)
    def put_language_strings(self, item: LanguageStrings, context: PipelineContext = None) -> None:
        self._put(LanguageStrings, item, uniquekeys.for_language_strings, context=context)

    @put_many.register(LanguageStrings)
    def put_many_language_strings(self, items: Iterable[LanguageStrings], context: PipelineContext = None) -> None:
        self._put_many(LanguageStrings, items, uniquekeys.for_language_strings, context=context)

    @get.register(LanguageStringsData)
    @validate_query(uniquekeys.validate_language_strings_query, uniquekeys.convert_region_to_platform)
    def get_language_strings_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> LanguageStringsData:
        result = self.get_language_strings(query=query, context=context)
        if result._data[LanguageStringsData] is not None and result._Ghost__is_loaded(LanguageStringsData):
            return result._data[LanguageStringsData]
        else:
            raise NotFoundError

    # Map

    @get.register(Map)
    @validate_query(uniquekeys.validate_map_query, uniquekeys.convert_region_to_platform)
    def get_map(self, query: Mapping[str, Any], context: PipelineContext = None) -> Map:
        return self._get(Map, query, uniquekeys.for_map_query, context)

    @get_many.register(Map)
    @validate_query(uniquekeys.validate_many_map_query, uniquekeys.convert_region_to_platform)
    def get_many_map(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Map, None, None]:
        return self._get_many(Map, query, uniquekeys.for_many_map_query, context)

    @put.register(Map)
    def put_map(self, item: Map, context: PipelineContext = None) -> None:
        self._put(Map, item, uniquekeys.for_map, context=context)

    @put_many.register(Map)
    def put_many_map(self, items: Iterable[Map], context: PipelineContext = None) -> None:
        self._put_many(Map, items, uniquekeys.for_map, context=context)

    @get.register(MapData)
    @validate_query(uniquekeys.validate_map_query, uniquekeys.convert_region_to_platform)
    def get_map_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> MapData:
        result = self.get_map(query=query, context=context)
        if result._data[MapData] is not None and result._Ghost__is_loaded(MapData):
            return result._data[MapData]
        else:
            raise NotFoundError

    # Maps

    @get.register(Maps)
    @validate_query(uniquekeys.validate_maps_query, uniquekeys.convert_region_to_platform)
    def get_maps(self, query: Mapping[str, Any], context: PipelineContext = None) -> Maps:
        return self._get(Maps, query, uniquekeys.for_maps_query, context)

    @get_many.register(Maps)
    @validate_query(uniquekeys.validate_many_maps_query, uniquekeys.convert_region_to_platform)
    def get_many_maps(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Maps, None, None]:
        return self._get_many(Maps, query, uniquekeys.for_many_maps_query, context)

    @put.register(Maps)
    def put_maps(self, item: Maps, context: PipelineContext = None) -> None:
        self._put(Maps, item, uniquekeys.for_maps, context=context)
        for map in item:
            self._put(Map, map, uniquekeys.for_map, context=context)

    @put_many.register(Maps)
    def put_many_maps(self, items: Iterable[Maps], context: PipelineContext = None) -> None:
        self._put_many(Maps, items, uniquekeys.for_maps, context=context)

    # Profile Icon

    @get.register(ProfileIcon)
    @validate_query(uniquekeys.validate_profile_icon_query, uniquekeys.convert_region_to_platform)
    def get_profile_icon(self, query: Mapping[str, Any], context: PipelineContext = None) -> ProfileIcon:
        return self._get(ProfileIcon, query, uniquekeys.for_profile_icon_query, context)

    @get_many.register(ProfileIcon)
    @validate_query(uniquekeys.validate_many_profile_icon_query, uniquekeys.convert_region_to_platform)
    def get_many_profile_icon(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[ProfileIcon, None, None]:
        return self._get_many(ProfileIcon, query, uniquekeys.for_many_profile_icon_query, context)

    @put.register(ProfileIcon)
    def put_profile_icon(self, item: ProfileIcon, context: PipelineContext = None) -> None:
        self._put(ProfileIcon, item, uniquekeys.for_profile_icon, context=context)

    @put_many.register(ProfileIcon)
    def put_many_profile_icon(self, items: Iterable[ProfileIcon], context: PipelineContext = None) -> None:
        self._put_many(ProfileIcon, items, uniquekeys.for_profile_icon, context=context)

    @get.register(ProfileIconData)
    @validate_query(uniquekeys.validate_profile_icon_query, uniquekeys.convert_region_to_platform)
    def get_profile_icon_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> ProfileIconData:
        result = self.get_profile_icon(query=query, context=context)
        if result._data[ProfileIconData] is not None and result._Ghost__is_loaded(ProfileIconData):
            return result._data[ProfileIconData]
        else:
            raise NotFoundError

    # Profile Icons

    @get.register(ProfileIcons)
    @validate_query(uniquekeys.validate_profile_icons_query, uniquekeys.convert_region_to_platform)
    def get_profile_icons(self, query: Mapping[str, Any], context: PipelineContext = None) -> ProfileIcons:
        return self._get(ProfileIcons, query, uniquekeys.for_profile_icons_query, context)

    @get_many.register(ProfileIcons)
    @validate_query(uniquekeys.validate_many_profile_icons_query, uniquekeys.convert_region_to_platform)
    def get_many_profile_icons(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[ProfileIcons, None, None]:
        return self._get_many(ProfileIcons, query, uniquekeys.for_many_profile_icons_query, context)

    @put.register(ProfileIcons)
    def put_profile_icons(self, item: ProfileIcons, context: PipelineContext = None) -> None:
        self._put(ProfileIcons, item, uniquekeys.for_profile_icons, context=context)
        for profile_icon in item:
            self._put(ProfileIcon, profile_icon, uniquekeys.for_profile_icon, context=context)

    @put_many.register(ProfileIcons)
    def put_many_profile_icons(self, items: Iterable[ProfileIcons], context: PipelineContext = None) -> None:
        self._put_many(ProfileIcons, items, uniquekeys.for_profile_icons, context=context)

    # Realm

    @get.register(Realms)
    @validate_query(uniquekeys.validate_realms_query, uniquekeys.convert_region_to_platform)
    def get_realms(self, query: Mapping[str, Any], context: PipelineContext = None) -> Realms:
        return self._get(Realms, query, uniquekeys.for_realms_query, context)

    @get_many.register(Realms)
    @validate_query(uniquekeys.validate_many_realms_query, uniquekeys.convert_region_to_platform)
    def get_many_realms(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Realms, None, None]:
        return self._get_many(Realms, query, uniquekeys.for_many_realms_query, context)

    @put.register(Realms)
    def put_realms(self, item: Realms, context: PipelineContext = None) -> None:
        self._put(Realms, item, uniquekeys.for_realms, context=context)

    @put_many.register(Realms)
    def put_many_realms(self, items: Iterable[Realms], context: PipelineContext = None) -> None:
        self._put_many(Realms, items, uniquekeys.for_realms, context=context)

    @get.register(RealmData)
    @validate_query(uniquekeys.validate_realms_query, uniquekeys.convert_region_to_platform)
    def get_realms_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> RealmData:
        result = self.get_realms(query=query, context=context)
        if result._data[RealmData] is not None and result._Ghost__is_loaded(RealmData):
            return result._data[RealmData]
        else:
            raise NotFoundError

    # Rune

    @get.register(Rune)
    @validate_query(uniquekeys.validate_rune_query, uniquekeys.convert_region_to_platform)
    def get_rune(self, query: Mapping[str, Any], context: PipelineContext = None) -> Rune:
        return self._get(Rune, query, uniquekeys.for_rune_query, context)

    @get_many.register(Rune)
    @validate_query(uniquekeys.validate_many_rune_query, uniquekeys.convert_region_to_platform)
    def get_many_rune(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Rune, None, None]:
        return self._get_many(Rune, query, uniquekeys.for_many_rune_query, context)

    @put.register(Rune)
    def put_rune(self, item: Rune, context: PipelineContext = None) -> None:
        self._put(Rune, item, uniquekeys.for_rune, context=context)

    @put_many.register(Rune)
    def put_many_rune(self, items: Iterable[Rune], context: PipelineContext = None) -> None:
        self._put_many(Rune, items, uniquekeys.for_rune, context=context)

    @get.register(RuneData)
    @validate_query(uniquekeys.validate_rune_query, uniquekeys.convert_region_to_platform)
    def get_rune_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> RuneData:
        result = self.get_rune(query=query, context=context)
        if result._data[RuneData] is not None and result._Ghost__is_loaded(RuneData):
            return result._data[RuneData]
        else:
            raise NotFoundError

    # Runes

    @get.register(Runes)
    @validate_query(uniquekeys.validate_runes_query, uniquekeys.convert_region_to_platform)
    def get_runes(self, query: Mapping[str, Any], context: PipelineContext = None) -> Runes:
        return self._get(Runes, query, uniquekeys.for_runes_query, context)

    @get_many.register(Runes)
    @validate_query(uniquekeys.validate_many_runes_query, uniquekeys.convert_region_to_platform)
    def get_many_runes(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Runes, None, None]:
        return self._get_many(Runes, query, uniquekeys.for_many_runes_query, context)

    @put.register(Runes)
    def put_runes(self, item: Runes, context: PipelineContext = None) -> None:
        self._put(Runes, item, uniquekeys.for_runes, context=context)
        for rune in item:
            self._put(Rune, rune, uniquekeys.for_rune, context=context)

    @put_many.register(Runes)
    def put_many_runes(self, items: Iterable[Runes], context: PipelineContext = None) -> None:
        self._put_many(Runes, items, uniquekeys.for_runes, context=context)

    # Summoner Spell

    @get.register(SummonerSpell)
    @validate_query(uniquekeys.validate_summoner_spell_query, uniquekeys.convert_region_to_platform)
    def get_summoner_spell(self, query: Mapping[str, Any], context: PipelineContext = None) -> SummonerSpell:
        return self._get(SummonerSpell, query, uniquekeys.for_summoner_spell_query, context)

    @get_many.register(SummonerSpell)
    @validate_query(uniquekeys.validate_many_summoner_spell_query, uniquekeys.convert_region_to_platform)
    def get_many_summoner_spell(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[SummonerSpell, None, None]:
        return self._get_many(SummonerSpell, query, uniquekeys.for_many_summoner_spell_query, context)

    @put.register(SummonerSpell)
    def put_summoner_spell(self, item: SummonerSpell, context: PipelineContext = None) -> None:
        self._put(SummonerSpell, item, uniquekeys.for_summoner_spell, context=context)

    @put_many.register(SummonerSpell)
    def put_many_summoner_spell(self, items: Iterable[SummonerSpell], context: PipelineContext = None) -> None:
        self._put_many(SummonerSpell, items, uniquekeys.for_summoner_spell, context=context)

    @get.register(SummonerSpellData)
    @validate_query(uniquekeys.validate_summoner_spell_query, uniquekeys.convert_region_to_platform)
    def get_summoner_spell_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> SummonerSpellData:
        result = self.get_summoner_spell(query=query, context=context)
        if result._data[SummonerSpellData] is not None and result._Ghost__is_loaded(SummonerSpellData):
            return result._data[SummonerSpellData]
        else:
            raise NotFoundError

    # Summoner Spells

    @get.register(SummonerSpells)
    @validate_query(uniquekeys.validate_summoner_spells_query, uniquekeys.convert_region_to_platform)
    def get_summoner_spells(self, query: Mapping[str, Any], context: PipelineContext = None) -> SummonerSpells:
        return self._get(SummonerSpells, query, uniquekeys.for_summoner_spells_query, context)

    @get_many.register(SummonerSpells)
    @validate_query(uniquekeys.validate_many_summoner_spells_query, uniquekeys.convert_region_to_platform)
    def get_many_summoner_spells(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[SummonerSpells, None, None]:
        return self._get_many(SummonerSpells, query, uniquekeys.for_many_summoner_spells_query, context)

    @put.register(SummonerSpells)
    def put_summoner_spells(self, item: SummonerSpells, context: PipelineContext = None) -> None:
        self._put(SummonerSpells, item, uniquekeys.for_summoner_spells, context=context)
        for summoner_spell in item:
            self._put(SummonerSpell, summoner_spell, uniquekeys.for_summoner_spell, context=context)

    @put_many.register(SummonerSpells)
    def put_many_summoner_spells(self, items: Iterable[SummonerSpells], context: PipelineContext = None) -> None:
        self._put_many(SummonerSpells, items, uniquekeys.for_summoner_spells, context=context)

    # Versions

    @get.register(Versions)
    @validate_query(uniquekeys.validate_versions_query, uniquekeys.convert_region_to_platform)
    def get_versions(self, query: Mapping[str, Any], context: PipelineContext = None) -> Versions:
        return self._get(Versions, query, uniquekeys.for_versions_query, context=context)

    @get_many.register(Versions)
    @validate_query(uniquekeys.validate_many_versions_query, uniquekeys.convert_region_to_platform)
    def get_many_versions(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Versions, None, None]:
        return self._get_many(Versions, query, uniquekeys.for_many_versions_query, context=context)

    @put.register(Versions)
    def put_versions(self, item: Versions, context: PipelineContext = None) -> None:
        self._put(Versions, item, uniquekeys.for_versions, context=context)

    @put_many.register(Versions)
    def put_many_versions(self, items: Iterable[Versions], context: PipelineContext = None) -> None:
        self._put_many(Versions, items, uniquekeys.for_versions, context=context)

    ##############
    # Status API #
    ##############

    @get.register(ShardStatus)
    @validate_query(uniquekeys.validate_shard_status_query, uniquekeys.convert_region_to_platform)
    def get_shard_status(self, query: Mapping[str, Any], context: PipelineContext = None) -> ShardStatus:
        return self._get(ShardStatus, query, uniquekeys.for_shard_status_query, context)

    @get_many.register(ShardStatus)
    @validate_query(uniquekeys.validate_many_shard_status_query, uniquekeys.convert_region_to_platform)
    def get_many_shard_status(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[ShardStatus, None, None]:
        return self._get_many(ShardStatus, query, uniquekeys.for_many_shard_status_query, context)

    @put.register(ShardStatus)
    def put_shard_status(self, item: ShardStatus, context: PipelineContext = None) -> None:
        self._put(ShardStatus, item, uniquekeys.for_shard_status, context=context)

    @put_many.register(ShardStatus)
    def put_many_shard_status(self, items: Iterable[ShardStatus], context: PipelineContext = None) -> None:
        self._put_many(ShardStatus, items, uniquekeys.for_shard_status, context=context)

    @get.register(ShardStatusData)
    @validate_query(uniquekeys.validate_shard_status_query, uniquekeys.convert_region_to_platform)
    def get_shard_status_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> ShardStatusData:
        result = self.get_shard_status(query=query, context=context)
        if result._data[ShardStatusData] is not None and result._Ghost__is_loaded(ShardStatusData):
            return result._data[ShardStatusData]
        else:
            raise NotFoundError

    #############
    # Match API #
    #############

    @get.register(Match)
    @validate_query(uniquekeys.validate_match_query, uniquekeys.convert_region_to_platform)
    def get_match(self, query: Mapping[str, Any], context: PipelineContext = None) -> Match:
        return self._get(Match, query, uniquekeys.for_match_query, context)

    @get_many.register(Match)
    @validate_query(uniquekeys.validate_many_match_query, uniquekeys.convert_region_to_platform)
    def get_many_match(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Match, None, None]:
        return self._get_many(Match, query, uniquekeys.for_many_match_query, context)

    @put.register(Match)
    def put_match(self, item: Match, context: PipelineContext = None) -> None:
        self._put(Match, item, uniquekeys.for_match, context=context)

    @put_many.register(Match)
    def put_many_match(self, items: Iterable[Match], context: PipelineContext = None) -> None:
        self._put_many(Match, items, uniquekeys.for_match, context=context)

    @get.register(MatchData)
    @validate_query(uniquekeys.validate_match_query, uniquekeys.convert_region_to_platform)
    def get_match_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> MatchData:
        result = self.get_match(query=query, context=context)
        if result._data[MatchData] is not None and result._Ghost__is_loaded(MatchData):
            return result._data[MatchData]
        else:
            raise NotFoundError

    # Timeline

    @get.register(Timeline)
    @validate_query(uniquekeys.validate_match_timeline_query, uniquekeys.convert_region_to_platform)
    def get_match_timeline(self, query: Mapping[str, Any], context: PipelineContext = None) -> Timeline:
        return self._get(Timeline, query, uniquekeys.for_match_query, context)

    @get_many.register(Timeline)
    @validate_query(uniquekeys.validate_many_match_timeline_query, uniquekeys.convert_region_to_platform)
    def get_many_match_timeline(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Timeline, None, None]:
        return self._get_many(Timeline, query, uniquekeys.for_many_match_query, context)

    @put.register(Timeline)
    def put_match_timeline(self, item: Timeline, context: PipelineContext = None) -> None:
        self._put(Timeline, item, uniquekeys.for_match_timeline, context=context)

    @put_many.register(Timeline)
    def put_many_match_timeline(self, items: Iterable[Timeline], context: PipelineContext = None) -> None:
        self._put_many(Timeline, items, uniquekeys.for_match_timeline, context=context)

    @get.register(TimelineData)
    @validate_query(uniquekeys.validate_match_timeline_query, uniquekeys.convert_region_to_platform)
    def get_match_timeline_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> TimelineData:
        result = self.get_match_timeline(query=query, context=context)
        if result._data[TimelineData] is not None and result._Ghost__is_loaded(TimelineData):
            return result._data[TimelineData]
        else:
            raise NotFoundError

    #################
    # Spectator API #
    #################

    # Current Match

    @get.register(CurrentMatch)
    @validate_query(uniquekeys.validate_current_match_query, uniquekeys.convert_region_to_platform)
    def get_current_match(self, query: Mapping[str, Any], context: PipelineContext = None) -> CurrentMatch:
        return self._get(CurrentMatch, query, uniquekeys.for_current_match_query, context)

    @get_many.register(CurrentMatch)
    @validate_query(uniquekeys.validate_many_current_match_query, uniquekeys.convert_region_to_platform)
    def get_many_current_match(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[CurrentMatch, None, None]:
        return self._get_many(CurrentMatch, query, uniquekeys.for_many_current_match_query, context)

    @put.register(CurrentMatch)
    def put_current_match(self, item: CurrentMatch, context: PipelineContext = None) -> None:
        self._put(CurrentMatch, item, uniquekeys.for_current_match, context=context)

    @put_many.register(CurrentMatch)
    def put_many_current_match(self, items: Iterable[CurrentMatch], context: PipelineContext = None) -> None:
        self._put_many(CurrentMatch, items, uniquekeys.for_current_match, context=context)

    @get.register(CurrentGameInfoData)
    @validate_query(uniquekeys.validate_current_match_query, uniquekeys.convert_region_to_platform)
    def get_current_match_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> CurrentGameInfoData:
        result = self.get_current_match(query=query, context=context)
        if result._data[CurrentGameInfoData] is not None and result._Ghost__is_loaded(CurrentGameInfoData):
            return result._data[CurrentGameInfoData]
        else:
            raise NotFoundError

    # Featured Matches

    @get.register(FeaturedMatches)
    @validate_query(uniquekeys.validate_featured_matches_query, uniquekeys.convert_region_to_platform)
    def get_featured_matches(self, query: Mapping[str, Any], context: PipelineContext = None) -> FeaturedMatches:
        return self._get(FeaturedMatches, query, uniquekeys.for_featured_matches_query, context)

    @get_many.register(FeaturedMatches)
    @validate_query(uniquekeys.validate_many_featured_matches_query, uniquekeys.convert_region_to_platform)
    def get_many_featured_matches(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[FeaturedMatches, None, None]:
        return self._get_many(FeaturedMatches, query, uniquekeys.for_many_featured_matches_query, context)

    @put.register(FeaturedMatches)
    def put_featured_matches(self, item: FeaturedMatches, context: PipelineContext = None) -> None:
        self._put(FeaturedMatches, item, uniquekeys.for_featured_matches, context=context)

    @put_many.register(FeaturedMatches)
    def put_many_featured_matches(self, items: Iterable[FeaturedMatches], context: PipelineContext = None) -> None:
        self._put_many(FeaturedMatches, items, uniquekeys.for_featured_matches, context=context)

    ################
    # Summoner API #
    ################

    @get.register(Summoner)
    @validate_query(uniquekeys.validate_summoner_query, uniquekeys.convert_region_to_platform)
    def get_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> Summoner:
        return self._get(Summoner, query, uniquekeys.for_summoner_query, context)

    @get_many.register(Summoner)
    @validate_query(uniquekeys.validate_many_summoner_query, uniquekeys.convert_region_to_platform)
    def get_many_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Summoner, None, None]:
        return self._get_many(Summoner, query, uniquekeys.for_many_summoner_query, context)

    @put.register(Summoner)
    def put_summoner(self, item: Summoner, context: PipelineContext = None) -> None:
        self._put(Summoner, item, uniquekeys.for_summoner, context=context)

    @put_many.register(Summoner)
    def put_many_summoner(self, items: Iterable[Summoner], context: PipelineContext = None) -> None:
        self._put_many(Summoner, items, uniquekeys.for_summoner, context=context)

    @get.register(SummonerData)
    @validate_query(uniquekeys.validate_summoner_query, uniquekeys.convert_region_to_platform)
    def get_summoner_data(self, query: Mapping[str, Any], context: PipelineContext = None) -> SummonerData:
        result = self.get_summoner(query=query, context=context)
        if result._data[SummonerData] is not None and result._Ghost__is_loaded(SummonerData):
            return result._data[SummonerData]
        else:
            raise NotFoundError



from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator

from datapipelines import DataSource, PipelineContext, Query, NotFoundError, validate_query
from .common import RiotAPIService, APINotFoundError
from ...data import Platform, Queue, Tier, Division
from ...dto.league import LeagueEntriesDto, LeagueDto, LeagueSummonerEntriesDto, ChallengerLeagueListDto, MasterLeagueListDto, GrandmasterLeagueListDto
from ..uniquekeys import convert_region_to_platform

T = TypeVar("T")


class LeaguesAPI(RiotAPIService):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    # League Entries

    _validate_get_league_entries_query = Query. \
        has("queue").as_(Queue).also. \
        has("tier").as_(Tier).also. \
        has("division").as_(Division).also. \
        has("page").as_(int).also. \
        has("platform").as_(Platform)

    @get.register(LeagueEntriesDto)
    @validate_query(_validate_get_league_entries_query, convert_region_to_platform)
    def get_league_entries_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LeagueEntriesDto:
        url = "https://{platform}.api.riotgames.com/lol/league/v4/entries/{queue}/{tier}/{division}".format(
            platform=query["platform"].value.lower(),
            queue=query["queue"].value,
            tier=query["tier"].value,
            division=query["division"].value
        )
        try:
            app_limiter, method_limiter = self._get_rate_limiter(query["platform"], "leagues/paginated-entries")
            data = self._get(url, parameters={"page": query["page"]}, app_limiter=app_limiter, method_limiter=method_limiter)
        except APINotFoundError:
            data = []
        region = query["platform"].region.value
        for entry in data:
            entry["region"] = region
        return LeagueEntriesDto(entries=data, page=query["page"], region=query["region"].value, queue=query["queue"].value, tier=query["tier"].value, division=query["division"].value)


    _validate_get_league_summoner_entries_query = Query. \
        has("summoner.id").as_(str).also. \
        has("platform").as_(Platform)

    @get.register(LeagueSummonerEntriesDto)
    @validate_query(_validate_get_league_summoner_entries_query, convert_region_to_platform)
    def get_league_summoner_entries_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LeagueSummonerEntriesDto:
        url = "https://{platform}.api.riotgames.com/lol/league/v4/entries/by-summoner/{id}".format(
            platform=query["platform"].value.lower(),
            id=query["summoner.id"]
        )
        try:
            app_limiter, method_limiter = self._get_rate_limiter(query["platform"], "leagues/summoner-entries")
            data = self._get(url, app_limiter=app_limiter, method_limiter=method_limiter)
        except APINotFoundError:
            data = []
        region = query["platform"].region.value
        for entry in data:
            entry["region"] = region
        return LeagueSummonerEntriesDto(entries=data, region=region, summonerId=query["summoner.id"])

    # Leagues

    _validate_get_league_query = Query. \
        has("id").as_(str).also. \
        has("platform").as_(Platform)

    @get.register(LeagueDto)
    @validate_query(_validate_get_league_query, convert_region_to_platform)
    def get_leagues_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LeagueDto:
        url = "https://{platform}.api.riotgames.com/lol/league/v4/leagues/{leagueId}".format(platform=query["platform"].value.lower(), leagueId=query["id"])
        try:
            endpoint = "leagues/leagueId {}".format(query["platform"].value)
            app_limiter, method_limiter = self._get_rate_limiter(query["platform"], endpoint)
            data = self._get(url, {}, app_limiter=app_limiter, method_limiter=method_limiter)
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        for entry in data["entries"]:
            entry["region"] = data["region"]
            entry["tier"] = data["tier"]
        return LeagueDto(data)

    _validate_get_many_league_query = Query. \
        has("ids").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(LeagueDto)
    @validate_query(_validate_get_many_league_query, convert_region_to_platform)
    def get_many_leagues_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[LeagueDto, None, None]:
        def generator():
            for id in query["ids"]:
                url = "https://{platform}.api.riotgames.com/lol/league/v4/leagues/{leagueId}".format(platform=query["platform"].value.lower(), leagueId=id)
                try:
                    endpoint = "leagues/leagueId {}".format(query["platform"].value)
                    app_limiter, method_limiter = self._get_rate_limiter(query["platform"], endpoint)
                    data = self._get(url, {}, app_limiter=app_limiter, method_limiter=method_limiter)
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data = {"leagues": data}
                data["region"] = query["platform"].region.value
                for league in data["leagues"]:
                    league["region"] = data["region"]
                    for entry in league["entries"]:
                        entry["region"] = data["region"]
                yield LeagueDto(data)

        return generator()

    _validate_get_challenger_league_query = Query. \
        has("queue").as_(Queue).also. \
        has("platform").as_(Platform)

    @get.register(ChallengerLeagueListDto)
    @validate_query(_validate_get_challenger_league_query, convert_region_to_platform)
    def get_challenger_league_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChallengerLeagueListDto:
        url = "https://{platform}.api.riotgames.com/lol/league/v4/challengerleagues/by-queue/{queueName}".format(platform=query["platform"].value.lower(), queueName=query["queue"].value)
        try:
            endpoint = "challengerleagues/by-queue {}".format(query["platform"].value)
            app_limiter, method_limiter = self._get_rate_limiter(query["platform"], endpoint)
            data = self._get(url, {}, app_limiter=app_limiter, method_limiter=method_limiter)
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["queue"] = query["queue"].value
        for entry in data["entries"]:
            entry["region"] = data["region"]
        return ChallengerLeagueListDto(data)

    _validate_get_many_challenger_league_query = Query. \
        has("queues").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(ChallengerLeagueListDto)
    @validate_query(_validate_get_many_challenger_league_query, convert_region_to_platform)
    def get_many_challenger_leagues_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ChallengerLeagueListDto, None, None]:
        def generator():
            for queue in query["queues"]:
                url = "https://{platform}.api.riotgames.com/lol/league/v4/challengerleagues/by-queue/{queueName}".format(platform=query["platform"].value.lower(), queueName=queue.value)
                try:
                    endpoint = "challengerleagues/by-queue {}".format(query["platform"].value)
                    app_limiter, method_limiter = self._get_rate_limiter(query["platform"], endpoint)
                    data = self._get(url, {}, app_limiter=app_limiter, method_limiter=method_limiter)
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data = {"leagues": data}
                data["region"] = query["platform"].region.value
                data["queue"] = queue.value
                for entry in data["entries"]:
                    entry["region"] = data["region"]
                yield ChallengerLeagueListDto(data)

        return generator()

    _validate_get_grandmaster_league_query = Query. \
        has("queue").as_(Queue).also. \
        has("platform").as_(Platform)

    @get.register(GrandmasterLeagueListDto)
    @validate_query(_validate_get_grandmaster_league_query, convert_region_to_platform)
    def get_grandmaster_league_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> GrandmasterLeagueListDto:
        url = "https://{platform}.api.riotgames.com/lol/league/v4/grandmasterleagues/by-queue/{queueName}".format(platform=query["platform"].value.lower(), queueName=query["queue"].value)
        try:
            endpoint = "grandmasterleagues/by-queue {}".format(query["platform"].value)
            app_limiter, method_limiter = self._get_rate_limiter(query["platform"], endpoint)
            data = self._get(url, {}, app_limiter=app_limiter, method_limiter=method_limiter)
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["queue"] = query["queue"].value
        for entry in data["entries"]:
            entry["region"] = data["region"]
        return GrandmasterLeagueListDto(data)

    _validate_get_many_grandmaster_league_query = Query. \
        has("queues").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(GrandmasterLeagueListDto)
    @validate_query(_validate_get_many_grandmaster_league_query, convert_region_to_platform)
    def get_many_grandmaster_leagues_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[GrandmasterLeagueListDto, None, None]:
        def generator():
            for queue in query["queues"]:
                url = "https://{platform}.api.riotgames.com/lol/league/v4/grandmasterleagues/by-queue/{queueName}".format(platform=query["platform"].value.lower(), queueName=queue.value)
                try:
                    endpoint = "grandmasterleagues/by-queue {}".format(query["platform"].value)
                    app_limiter, method_limiter = self._get_rate_limiter(query["platform"], endpoint)
                    data = self._get(url, {}, app_limiter=app_limiter, method_limiter=method_limiter)
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data = {"leagues": data}
                data["region"] = query["platform"].region.value
                data["queue"] = queue.value
                for entry in data["entries"]:
                    entry["region"] = data["region"]
                yield GrandmasterLeagueListDto(data)

        return generator()

    _validate_get_master_league_query = Query. \
        has("queue").as_(Queue).also. \
        has("platform").as_(Platform)

    @get.register(MasterLeagueListDto)
    @validate_query(_validate_get_master_league_query, convert_region_to_platform)
    def get_master_league_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MasterLeagueListDto:
        url = "https://{platform}.api.riotgames.com/lol/league/v4/masterleagues/by-queue/{queueName}".format(platform=query["platform"].value.lower(), queueName=query["queue"].value)
        try:
            endpoint = "masterleagues/by-queue {}".format(query["platform"].value)
            app_limiter, method_limiter = self._get_rate_limiter(query["platform"], endpoint)
            data = self._get(url, {}, app_limiter=app_limiter, method_limiter=method_limiter)
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["queue"] = query["queue"].value
        for entry in data["entries"]:
            entry["region"] = data["region"]
        return MasterLeagueListDto(data)

    _validate_get_many_master_league_query = Query. \
        has("queues").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(MasterLeagueListDto)
    @validate_query(_validate_get_many_master_league_query, convert_region_to_platform)
    def get_many_master_leagues_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[MasterLeagueListDto, None, None]:
        def generator():
            for queue in query["queues"]:
                url = "https://{platform}.api.riotgames.com/lol/league/v4/masterleagues/by-queue/{queueName}".format(platform=query["platform"].value.lower(), queueName=queue.value)
                try:
                    endpoint = "masterleagues/by-queue {}".format(query["platform"].value)
                    app_limiter, method_limiter = self._get_rate_limiter(query["platform"], endpoint)
                    data = self._get(url, {}, app_limiter=app_limiter, method_limiter=method_limiter)
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data = {"leagues": data}
                data["region"] = query["platform"].region.value
                data["queue"] = queue.value
                for entry in data["entries"]:
                    entry["region"] = data["region"]
                yield MasterLeagueListDto(data)

        return generator()



from typing import List, Union

from merakicommons.cache import lazy
from merakicommons.container import searchable, SearchableList

from ..data import Region, Platform
from .common import CoreData, CassiopeiaObject, CassiopeiaGhost, provide_default_region, ghost_load_on


##############
# Data Types #
##############


class TranslationData(CoreData):
    _renamed = {"updated_at": "updated"}


class MessageData(CoreData):
    _renamed = {"created_at": "created", "updated_at": "updated"}

    def __call__(self, **kwargs):
        if "translations" in kwargs:
            self.translations = [TranslationData(**translation) for translation in kwargs.pop("translations")]
        super().__call__(**kwargs)
        return self


class IncidentData(CoreData):
    _renamed = {"created_at": "created"}

    def __call__(self, **kwargs):
        if "updates" in kwargs:
            self.updates = [MessageData(**update) for update in kwargs.pop("updates")]
        super().__call__(**kwargs)
        return self


class ServiceData(CoreData):
    _renamed = {}

    def __call__(self, **kwargs):
        if "incidents" in kwargs:
            self.incidents = [IncidentData(**incident) for incident in kwargs.pop("incidents")]
        super().__call__(**kwargs)
        return self


class ShardStatusData(CoreData):
    _renamed = {"region_tag": "platform"}

    def __call__(self, **kwargs):
        if "services" in kwargs:
            self.services = [ServiceData(**service) for service in kwargs.pop("services")]
        super().__call__(**kwargs)
        return self


##############
# Core Types #
##############


@searchable({})
class Translation(CassiopeiaObject):
    _data_types = {TranslationData}

    @property
    def locale(self) -> str:
        return self._data[TranslationData].locale

    @property
    def content(self) -> str:
        return self._data[TranslationData].content

    @property
    def updated(self) -> str:
        return self._data[TranslationData].updated


class Message(CassiopeiaObject):
    _data_types = {MessageData}

    @property
    def severity(self) -> str:
        return self._data[MessageData].severity

    @property
    def author(self) -> str:
        return self._data[MessageData].author

    @property
    def created(self) -> str:
        return self._data[MessageData].created

    @property
    def translations(self) -> List[Translation]:
        return SearchableList([Translation(trans) for trans in self._data[MessageData].translations])

    @property
    def updated(self) -> str:
        return self._data[MessageData].updated

    @property
    def content(self) -> str:
        return self._data[MessageData].content

    @property
    def id(self) -> str:
        return self._data[MessageData].id


class Incident(CassiopeiaObject):
    _data_types = {IncidentData}

    @property
    def active(self) -> bool:
        return self._data[IncidentData].active

    @property
    def created(self) -> str:
        return self._data[IncidentData].created

    @property
    def id(self) -> int:
        return self._data[IncidentData].id

    @property
    def updates(self) -> List[Message]:
        return SearchableList([Message.from_data(message) for message in self._data[IncidentData].updates])


class Service(CassiopeiaObject):
    _data_types = {ServiceData}

    @property
    def status(self) -> str:
        return self._data[ServiceData].status

    @property
    def incidents(self) -> List[Incident]:
        return SearchableList([Incident.from_data(inc) for inc in self._data[ServiceData].incidents])

    @property
    def name(self) -> str:
        return self._data[ServiceData].name

    @property
    def slug(self) -> str:
        return self._data[ServiceData].slug


@searchable({})
class ShardStatus(CassiopeiaGhost):
    _data_types = {ShardStatusData}

    @provide_default_region
    def __init__(self, region: Union[Region, str] = None):
        kwargs = {"region": region}
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform}

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on
    @lazy
    def region(self) -> Region:
        return Region(self._data[ShardStatusData].region)

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on
    @lazy
    def platform(self) -> Platform:
        return self.region.platform

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on
    def name(self) -> str:
        return self._data[ShardStatusData].name

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on
    def hostname(self) -> str:
        return self._data[ShardStatusData].hostname

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on
    def services(self) -> List[Service]:
        return SearchableList([Service.from_data(service) for service in self._data[ShardStatusData].services])

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on
    def slug(self) -> str:
        return self._data[ShardStatusData].slug

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on
    def locales(self) -> List[str]:
        return self._data[ShardStatusData].locales



from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator

from datapipelines import DataSource, PipelineContext, Query, NotFoundError, validate_query
from .common import KernelSource, APINotFoundError
from ...data import Platform
from ...dto.thirdpartycode import VerificationStringDto
from ..uniquekeys import convert_region_to_platform

T = TypeVar("T")


def _get_default_locale(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    return query["platform"].default_locale


class ThirdPartyCodeAPI(KernelSource):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    #######################
    # Verification String #
    #######################

    _validate_get_verification_string_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(str)

    @get.register(VerificationStringDto)
    @validate_query(_validate_get_verification_string_query, convert_region_to_platform)
    def get_verification_string(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> VerificationStringDto:
        parameters = {"platform": query["platform"].value}
        endpoint = "lol/platform/v4/third-party-code/by-summoner/{summonerId}".format(summonerId=query["summoner.id"])
        try:
            data = self._get(endpoint=endpoint, parameters=parameters)
        except (ValueError, APINotFoundError) as error:
            raise NotFoundError(str(error)) from error

        data = {"string": data}
        data["region"] = query["platform"].region.value
        data["summonerId"] = query["summoner.id"]
        return VerificationStringDto(data)

    _validate_get_many_verification_string_query = Query. \
        has("platforms").as_(Iterable).also. \
        has("summoner.ids").as_(Iterable)

    @get_many.register(VerificationStringDto)
    @validate_query(_validate_get_many_verification_string_query, convert_region_to_platform)
    def get_many_verification_string(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[VerificationStringDto, None, None]:
        def generator():
            parameters = {"platform": query["platform"].value}
            for platform, summoner_id in zip(query["platforms"], query["summoner.ids"]):
                platform = Platform(platform.upper())
                endpoint = "lol/platform/v4/third-party-code/by-summoner/{summonerId}".format(summonerId=summoner_id)
                try:
                    data = self._get(endpoint=endpoint, parameters=parameters)
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data = {"string": data}
                data["region"] = platform.region.value
                data["summonerId"] = summoner_id
                yield VerificationStringDto(data)

        return generator()



from typing import TypeVar, Type, Dict, Union, List, Mapping
import logging
import importlib
import inspect
import copy

from datapipelines import DataPipeline, DataSink, DataSource, CompositeDataTransformer, DataTransformer

from ..data import Region, Platform

T = TypeVar("T")

logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%Y-%m-%d %H:%M:%S', level=logging.WARNING)

def create_pipeline(service_configs: Dict, verbose: int = 0) -> DataPipeline:
    transformers = []

    # Always use the Riot API transformers
    from ..transformers import __transformers__ as riotapi_transformer
    transformers.extend(riotapi_transformer)

    # Add sources / sinks by name from config
    services = []
    for store_name, config in service_configs.items():
        try:
            package = config.pop("package", "_cassiopeia.datastores")
        except TypeError:
            package = "_cassiopeia.datastores"
        module = importlib.import_module(name=package)
        store_cls = getattr(module, store_name)
        if store_name.lower() == "djangocache":
            for cache_config in config:
                services.append(store_cls(**cache_config))
        else:
            store = store_cls(**config)
            services.append(store)
            service_transformers = getattr(module, "__transformers__", [])
            transformers.extend(service_transformers)

    from ..datastores import Cache, MerakiAnalyticsCDN, LolWikia

    # Automatically insert the ghost store if it isn't there
    from ..datastores import UnloadedGhostStore
    found = False
    for datastore in services:
        if isinstance(datastore, UnloadedGhostStore):
            found = True
            break
    if not found:
        if any(isinstance(service, Cache) for service in services):
            # Find the cache and insert the ghost store directly after it
            for i, datastore in enumerate(services):
                if isinstance(datastore, Cache):
                    services.insert(i+1, UnloadedGhostStore())
                    break
        else:
            # Insert the ghost store at the beginning of the pipeline
            services.insert(0, UnloadedGhostStore())

    services.append(MerakiAnalyticsCDN())
    services.append(LolWikia())
    pipeline = DataPipeline(services, transformers)

    if verbose > 0:
        for service in services:
            print("Service:", service)
            if verbose > 1:
                if isinstance(service, DataSource):
                    for p in service.provides:
                        print("  Provides:", p)
                if isinstance(service, DataSink):
                    for p in service.accepts:
                        print("  Accepts:", p)
        if verbose > 2:
            for transformer in transformers:
                for t in transformer.transforms.items():
                    print("Transformer:", t)
            print()

    return pipeline


def register_transformer_conversion(transformer: DataTransformer, from_type, to_type):
    # Find the method that takes a `from_type` and returns a `to_type` and register it
    methods = inspect.getmembers(transformer, predicate=inspect.ismethod)
    for name, method in methods:
        annotations = copy.copy(method.__annotations__)
        return_type = annotations.pop("return")
        annotations.pop("context", None)
        try:
            if to_type is return_type and from_type in annotations.values():
                transformer.transform.register(from_type, to_type)(method.__func__).__get__(transformer, transformer.__class__)
                break
        except TypeError:
            continue
    else:
        raise RuntimeError("Could not find method to register: {} to {} in {}.".format(from_type, to_type, transformer))


def get_default_config():
    return {
        "global": {
            "version_from_match": "patch",
            "default_region": None
        },
        "plugins": {},
        "pipeline": {
            "Cache": {},
            "DjangoCache": {},
            "DDragon": {},
            "RiotAPI": {
                "api_key": "RIOT_API_KEY"
            }
        },
        "logging": {
            "print_calls": True,
            "print_riot_api_key": False,
            "default": "WARNING",
            "core": "WARNING"
        }
    }


class Settings(object):
    def __init__(self, settings):
        _defaults = get_default_config()
        globals_ = settings.get("global", _defaults["global"])
        self.__version_from_match = globals_.get("version_from_match", _defaults["global"]["version_from_match"])  # Valid json values are: "version", "patch", and null
        self.__default_region = globals_.get("default_region", _defaults["global"]["default_region"])
        if self.__default_region is not None:
            self.__default_region = Region(self.__default_region.upper())

        self.__plugins = settings.get("plugins", _defaults["plugins"])

        self.__pipeline_args = settings.get("pipeline", _defaults["pipeline"])
        self.__pipeline = None  # type: DataPipeline

        logging_config = settings.get("logging", _defaults["logging"])
        self.__default_print_calls = logging_config.get("print_calls", _defaults["logging"]["print_calls"])
        self.__default_print_riot_api_key = logging_config.get("print_riot_api_key", _defaults["logging"]["print_riot_api_key"])
        for name in ["default", "core"]:
            logger = logging.getLogger(name)
            level = logging_config.get(name, _defaults["logging"][name])
            logger.setLevel(level)
            for handler in logger.handlers:
                handler.setLevel(level)

    def set_region(self, region: Union[Region, str]):
        if isinstance(region, str):
            region = Region(region.upper())
        self.__default_region = region

    @property
    def pipeline(self) -> DataPipeline:
        if self.__pipeline is None:
            self.__pipeline = create_pipeline(service_configs=self.__pipeline_args, verbose=0)
        return self.__pipeline

    @property
    def default_region(self):
        return self.__default_region

    @property
    def default_platform(self):
        return Platform[self.__default_region.name]

    @property
    def version_from_match(self):
        return self.__version_from_match

    @property
    def plugins(self):
        return self.__plugins

    def set_riot_api_key(self, key):
        from ..datastores.riotapi import RiotAPI
        for sources in self.pipeline._sources:
            for source in sources:
                if isinstance(source, RiotAPI):
                    source.set_api_key(key)

    def clear_sinks(self, type: Type[T] = None):
        types = {type}
        if type is not None:
            from ..core.common import CoreData, CassiopeiaObject
            if issubclass(type, CassiopeiaObject):
                for t in type._data_types:
                    types.add(t)
                    types.add(t._dto_type)
            elif issubclass(type, CoreData):
                types.add(type._dto_type)

        for sink in self.pipeline._sinks:
            for type in types:
                sink.clear(type)

    def expire_sinks(self, type: Type[T] = None):
        types = {type}
        if type is not None:
            from ..core.common import CoreData, CassiopeiaObject
            if issubclass(type, CassiopeiaObject):
                for t in type._data_types:
                    types.add(t)
                    types.add(t._dto_type)
            elif issubclass(type, CoreData):
                types.add(type._dto_type)

        for sink in self.pipeline._sinks:
            for type in types:
                sink.expire(type)


from typing import List, Set, Mapping, Union

from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList

from ...data import Region, Platform
from ..common import CoreData, CassiopeiaObject, CassiopeiaGhost, CoreDataList, get_latest_version, CassiopeiaLazyList, provide_default_region, ghost_load_on
from .common import ImageData, Sprite, Image
from .map import Map
from ...dto.staticdata import item as dto


##############
# Data Types #
##############


class ItemListData(CoreDataList):
    _dto_type = dto.ItemListDto
    _renamed = {"included_data": "includedData"}


class ItemTreeData(CoreData):
    _renamed = {}


class ItemStatsData(CoreData):
    _renamed = {"PercentCritDamageMod": "percentCriticalStrikeDamage", "PercentSpellBlockMod": "percentMagicResist", "PercentHPRegenMod": "percentHealthRegen", "PercentMovementSpeedMod": "percentMovespeed", "FlatSpellBlockMod": "magicResist", "FlatCritDamageMod": "criticalStrikeDamage", "FlatEnergyPoolMod": "energy", "PercentLifeStealMod": "lifeSteal", "FlatMPPoolMod": "mana", "FlatMovementSpeedMod": "movespeed", "PercentAttackSpeedMod": "percentAttackSpeed", "FlatBlockMod": "block", "PercentBlockMod": "percentBlock", "FlatEnergyRegenMod": "energyRegen", "PercentSpellVampMod": "spellVamp", "FlatMPRegenMod": "manaRegen", "PercentDodgeMod": "dodge", "FlatAttackSpeedMod": "attackSpeed", "FlatArmorMod": "armor", "FlatHPRegenMod": "healthRegen", "PercentMagicDamageMod": "percentAbilityPower", "PercentMPPoolMod": "percentMana", "FlatMagicDamageMod": "abilityPower", "PercentMPRegenMod": "percentManaRegen", "PercentPhysicalDamageMod": "percentAttackDamage", "FlatPhysicalDamageMod": "attackDamage", "PercentHPPoolMod": "percentHealth", "PercentArmorMod": "percentArmor", "PercentEXPBonus": "percentExpBonus", "FlatHPPoolMod": "health", "FlatCritChanceMod": "criticalStrikeChance", "FlatEXPBonus": "expBonus"}

    def __call__(self, **kwargs):
        if "flatCritChanceMode" in kwargs and "percentCritChanceMod" in kwargs:
            self.critical_strike_chance =  kwargs.pop("flatCritChanceMod") + kwargs.pop("percentCritChanceMod")
        super().__call__(**kwargs)
        return self


class GoldData(CoreData):
    _renamed = {}


class ItemData(CoreData):
    _dto_type = dto.ItemDto
    _renamed = {"hideFromAll": "hide", "colloq": "keywords", "requiredChampion": "champion", "depth": "tier", "stacks": "max_stacks", "included_data": "includedData"}

    def __call__(self, **kwargs):
        if "image" in kwargs:
            self.image = ImageData(**kwargs.pop("image"))
        if "gold" in kwargs:
            self.gold = GoldData(**kwargs.pop("gold"))
        if "into" in kwargs:
            self.buildsInto = [int(x) for x in kwargs.pop("into")]
        if "from" in kwargs:
            self.buildsFrom = [int(x) for x in kwargs.pop("from")]
        if "stats" in kwargs:
            self.stats = ItemStatsData(**kwargs.pop("stats"))
        if "colloq" in kwargs:
            self.keywords = set(kw for kw in kwargs.pop("colloq").split(";") if kw != "")
        if "maps" in kwargs:
            """List of maps where this item is available."""
            self.maps = [int(m) for m, tf in kwargs.pop("maps").items() if tf]
        super().__call__(**kwargs)
        return self


##############
# Core Types #
##############


class Items(CassiopeiaLazyList):
    _data_types = {ItemListData}

    @provide_default_region
    def __init__(self, *, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if included_data is None:
            included_data = {"all"}
        if locale is None and region is not None:
            locale = Region(region).default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if version:
            kwargs["version"] = version
        CassiopeiaObject.__init__(self, **kwargs)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[ItemListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def version(self) -> str:
        try:
            return self._data[ItemListData].version
        except AttributeError:
            version = get_latest_version(region=self.region, endpoint="item")
            self(version=version)
            return self._data[ItemListData].version

    @property
    def locale(self) -> str:
        """The locale for this item."""
        return self._data[ItemListData].locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additional information for this item when it's loaded."""
        return self._data[ItemListData].includedData


class ItemStats(CassiopeiaObject):
    _data_types = {ItemStatsData}

    @property
    def percent_critical_strike_damage(self) -> float:
        return self._data[ItemStatsData].percentCriticalStrikeDamage

    @property
    def percent_magic_resist(self) -> float:
        return self._data[ItemStatsData].percentMagicResist

    @property
    def percent_health_regen(self) -> float:
        return self._data[ItemStatsData].percentHealthRegen

    @property
    def percent_movespeed(self) -> float:
        return self._data[ItemStatsData].percentMovespeed

    @property
    def magic_resist(self) -> float:
        return self._data[ItemStatsData].magicResist

    @property
    def critical_strike_damage(self) -> float:
        return self._data[ItemStatsData].criticalStrikeDamage

    @property
    def energy(self) -> float:
        return self._data[ItemStatsData].energy

    @property
    def life_steal(self) -> float:
        return self._data[ItemStatsData].lifeSteal

    @property
    def mana(self) -> float:
        return self._data[ItemStatsData].mana

    @property
    def movespeed(self) -> float:
        return self._data[ItemStatsData].movespeed

    @property
    def percent_attack_speed(self) -> float:
        return self._data[ItemStatsData].percentAttackSpeed

    @property
    def block(self) -> float:
        return self._data[ItemStatsData].block

    @property
    def percent_block(self) -> float:
        return self._data[ItemStatsData].percentBlock

    @property
    def energy_regen(self) -> float:
        return self._data[ItemStatsData].energyRegen

    @property
    def spell_vamp(self) -> float:
        return self._data[ItemStatsData].spellVamp

    @property
    def mana_regen(self) -> float:
        return self._data[ItemStatsData].manaRegen

    @property
    def dodge(self) -> float:
        return self._data[ItemStatsData].dodge

    @property
    def attack_speed(self) -> float:
        return self._data[ItemStatsData].attackSpeed

    @property
    def armor(self) -> float:
        return self._data[ItemStatsData].armor

    @property
    def health_regen(self) -> float:
        return self._data[ItemStatsData].healthRegen

    @property
    def percent_ability_power(self) -> float:
        return self._data[ItemStatsData].percentAbilityPower

    @property
    def percent_mana_regen(self) -> float:
        return self._data[ItemStatsData].percentManaRegen

    @property
    def ability_power(self) -> float:
        return self._data[ItemStatsData].abilityPower

    @property
    def percent_mana_regen(self) -> float:
        return self._data[ItemStatsData].percentManaRegen

    @property
    def percent_attack_damage(self) -> float:
        return self._data[ItemStatsData].percentAttackDamage

    @property
    def attack_damage(self) -> float:
        return self._data[ItemStatsData].attackDamage

    @property
    def percent_health(self) -> float:
        return self._data[ItemStatsData].percentHealth

    @property
    def percent_armor(self) -> float:
        return self._data[ItemStatsData].percentArmor

    @property
    def percent_xp_bonus(self) -> float:
        return self._data[ItemStatsData].percentExpBonus

    @property
    def health(self) -> float:
        return self._data[ItemStatsData].health

    @property
    def critical_strike_chance(self) -> float:
        return self._data[ItemStatsData].criticalStrikeChance

    @property
    def xp_bonus(self) -> float:
        return self._data[ItemStatsData].expBonus


class Gold(CassiopeiaObject):
    _data_types = {GoldData}

    @property
    def sell(self) -> int:
        return self._data[GoldData].sell

    @property
    def total(self) -> int:
        return self._data[GoldData].total

    @property
    def base(self) -> int:
        return self._data[GoldData].base

    @property
    def purchasable(self) -> bool:
        return self._data[GoldData].purchasable


@searchable({str: ["name", "region", "platform", "locale", "keywords", "maps", "tags", "tier"], int: ["id"], Region: ["region"], Platform: ["platform"], Map: ["maps"]})
class Item(CassiopeiaGhost):
    _data_types = {ItemData}

    @provide_default_region
    def __init__(self, *, id: int = None, name: str = None, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if included_data is None:
            included_data = {"all"}
        if locale is None and region is not None:
            locale = Region(region).default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if id is not None:
            kwargs["id"] = id
        if name is not None:
            kwargs["name"] = name
        if version is not None:
            kwargs["version"] = version
        super().__init__(**kwargs)

    def __get_query__(self):
        query = {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}
        if hasattr(self._data[ItemData], "id"):
            query["id"] = self._data[ItemData].id
        if hasattr(self._data[ItemData], "name"):
            query["name"] = self._data[ItemData].name
        return query

    def __eq__(self, other: "Item"):
        if not isinstance(other, Item) or self.region != other.region:
            return False
        s = {}
        o = {}
        if hasattr(self._data[ItemData], "id"): s["id"] = self.id
        if hasattr(other._data[ItemData], "id"): o["id"] = other.id
        if hasattr(self._data[ItemData], "name"): s["name"] = self.name
        if hasattr(other._data[ItemData], "name"): o["name"] = other.name
        if any(s.get(key, "s") == o.get(key, "o") for key in s):
            return True
        else:
            return self.id == other.id

    def __str__(self):
        region = self.region
        id_ = "?"
        name = "?"
        if hasattr(self._data[ItemData], "id"):
            id_ = self.id
        if hasattr(self._data[ItemData], "name"):
            name = self.name
        return "Item(name='{name}', id={id_}, region='{region}')".format(name=name, id_=id_, region=region.value)

    __hash__ = CassiopeiaGhost.__hash__

    @lazy_property
    def region(self) -> Region:
        """The region for this item."""
        return Region(self._data[ItemData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this item."""
        return self.region.platform

    @property
    def version(self) -> str:
        """The version for this item."""
        try:
            return self._data[ItemData].version
        except AttributeError:
            version = get_latest_version(region=self.region, endpoint="item")
            self(version=version)
            return self._data[ItemData].version

    @property
    def locale(self) -> str:
        """The locale for this item."""
        return self._data[ItemData].locale or self.region.default_locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additonal information for this item when it's loaded."""
        return self._data[ItemData].includedData

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def id(self) -> int:
        """The item's ID."""
        return self._data[ItemData].id

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def gold(self) -> Gold:
        return Gold.from_data(self._data[ItemData].gold)

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def plaintext(self) -> str:
        return self._data[ItemData].plaintext

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def hide(self) -> bool:
        return self._data[ItemData].hide

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def in_store(self) -> bool:
        return self._data[ItemData].in_store

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def builds_into(self) -> List["Item"]:
        if hasattr(self._data[ItemData], "buildsInto"):
            return SearchableList([Item(id=id_, region=self.region) for id_ in self._data[ItemData].buildsInto])
        else:
            return SearchableList([])

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def builds_from(self) -> List["Item"]:
        if hasattr(self._data[ItemData], "buildsFrom"):
            return SearchableList([Item(id=id_, region=self.region) for id_ in self._data[ItemData].buildsFrom])
        else:
            return SearchableList([])

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def stats(self) -> ItemStats:
        return self._data[ItemData].stats

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def keywords(self) -> List[str]:
        return SearchableList(self._data[ItemData].keywords)

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    @lazy
    def maps(self) -> List[Map]:
        return [Map(id=id_, region=self.region, version=self.version) for id_ in self._data[ItemData].maps]

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def description(self) -> str:
        return self._data[ItemData].description

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def tags(self) -> List[str]:
        return self._data[ItemData].tags

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def effect(self) -> Mapping[str, str]:
        return self._data[ItemData].effect

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def champion(self) -> "Champion":
        from .champion import Champion
        return Champion(name=self._data[ItemData].champion, region=self.region, version=self.verion)

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def group(self) -> str:
        return self._data[ItemData].group

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def name(self) -> str:
        return self._data[ItemData].name

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def consume_on_full(self) -> bool:
        return self._data[ItemData].consume_on_full

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def consumed(self) -> bool:
        return self._data[ItemData].consumed

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def sanitized_description(self) -> str:
        return self._data[ItemData].sanitized_description

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def tier(self) -> int:
        return self._data[ItemData].tier

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def max_stacks(self) -> int:
        return self._data[ItemData].max_stacks

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    def special_recipe(self) -> int:
        return self._data[ItemData].special_recipe

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on
    @lazy
    def image(self) -> Image:
        """The image information for this item."""
        image = Image.from_data(self._data[ItemData].image)
        image(version=self.version)
        return image

    @lazy_property
    def sprite(self) -> Sprite:
        return self.image.sprite_info
